from abc import abstractmethod

from foxes.utils import new_instance
from .model import Model


class AxialInductionModel(Model):
    """
    Abstract base class for axial induction models

    :group: core

    """

    @abstractmethod
    def ct2a(self, ct):
        """
        Computes induction from ct

        Parameters
        ----------
        ct: numpy.ndarray or float
            The ct values

        Returns
        -------
        ct: numpy.ndarray or float
            The induction values

        """
        pass

    @classmethod
    def new(cls, induction_type, *args, **kwargs):
        """
        Run-time axial induction model factory.

        Parameters
        ----------
        induction_type: str
            The selected derived class name
        args: tuple, optional
            Additional parameters for the constructor
        kwargs: dict, optional
            Additional parameters for the constructor

        """
        return new_instance(cls, induction_type, *args, **kwargs)



from abc import abstractmethod

from foxes.utils import new_instance

from .model import Model


class WakeSuperposition(Model):
    """
    Abstract base class for wake superposition models.

    Note that it is a matter of the wake model
    if superposition models are used, or if the
    wake model computes the total wake result by
    other means.

    :group: core

    """

    @abstractmethod
    def add_wake(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        st_sel,
        variable,
        wake_delta,
        wake_model_result,
    ):
        """
        Add a wake delta to previous wake deltas,
        at rotor points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order
        st_sel: numpy.ndarray of bool
            The selection of targets, shape: (n_states, n_targets)
        variable: str
            The variable name for which the wake deltas applies
        wake_delta: numpy.ndarray
            The original wake deltas, shape:
            (n_states, n_targets, n_tpoints, ...)
        wake_model_result: numpy.ndarray
            The new wake deltas of the selected rotors,
            shape: (n_st_sel, n_tpoints, ...)

        Returns
        -------
        wdelta: numpy.ndarray
            The updated wake deltas, shape:
            (n_states, n_targets, n_tpoints, ...)

        """
        pass

    @abstractmethod
    def calc_final_wake_delta(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        variable,
        wake_delta,
    ):
        """
        Calculate the final wake delta after adding all
        contributions.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        variable: str
            The variable name for which the wake deltas applies
        wake_delta: numpy.ndarray
            The wake deltas at targets, shape:
            (n_states, n_targets, n_tpoints)

        Returns
        -------
        final_wake_delta: numpy.ndarray
            The final wake delta, which will be added to the ambient
            results by simple plus operation. Shape:
            (n_states, n_targets, n_tpoints)

        """
        pass

    @classmethod
    def new(cls, superp_type, *args, **kwargs):
        """
        Run-time wake superposition model factory.

        Parameters
        ----------
        superp_type: str
            The selected derived class name
        args: tuple, optional
            Additional parameters for constructor
        kwargs: dict, optional
            Additional parameters for constructor

        """
        return new_instance(cls, superp_type, *args, **kwargs)


class WindVectorWakeSuperposition(Model):
    """
    Base class for wind vector superposition.

    Note that it is a matter of the wake model
    if superposition models are used, or if the
    wake model computes the total wake result by
    other means.

    :group: core

    """

    @abstractmethod
    def add_wake_vector(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
        st_sel,
        wake_delta_uv,
        wake_model_result_uv,
    ):
        """
        Add a wake delta vector to previous wake deltas,
        at rotor points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order
        st_sel: numpy.ndarray of bool
            The selection of targets, shape: (n_states, n_targets)
        wake_delta_uv: numpy.ndarray
            The original wind vector wake deltas, shape:
            (n_states, n_targets, n_tpoints, 2)
        wake_model_result_uv: numpy.ndarray
            The new wind vector wake deltas of the selected rotors,
            shape: (n_st_sel, n_tpoints, 2, ...)

        Returns
        -------
        wdelta_uv: numpy.ndarray
            The updated wind vector wake deltas, shape:
            (n_states, n_targets, n_tpoints, ...)

        """
        pass

    @abstractmethod
    def calc_final_wake_delta_uv(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        wake_delta_uv,
    ):
        """
        Calculate the final wind vector wake delta after adding all
        contributions.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        wake_delta_uv: numpy.ndarray
            The original wind vector wake deltas, shape:
            (n_states, n_targets, n_tpoints, 2)

        Returns
        -------
        final_wake_delta_ws: numpy.ndarray
            The final wind speed wake delta, which will be added to
            the ambient results by simple plus operation. Shape:
            (n_states, n_targets, n_tpoints)
        final_wake_delta_wd: numpy.ndarray
            The final wind direction wake delta, which will be added to
            the ambient results by simple plus operation. Shape:
            (n_states, n_targets, n_tpoints)

        """
        pass

    @classmethod
    def new(cls, superp_type, *args, **kwargs):
        """
        Run-time wind wake superposition model factory.

        Parameters
        ----------
        superp_type: str
            The selected derived class name
        args: tuple, optional
            Additional parameters for constructor
        kwargs: dict, optional
            Additional parameters for constructor

        """
        return new_instance(cls, superp_type, *args, **kwargs)



from foxes.algorithms.sequential import SequentialPlugin
from foxes.models.wake_frames.seq_dynamic_wakes import SeqDynamicWakes


class SeqWakeDebugPlugin(SequentialPlugin):
    """
    Plugin for creating wake debug plots in animations

    Attributes
    ----------
    show_p: bool
        Flag for showing wake points
    show_v: bool
        Flag for showing wake vectors
    vpars: dict
        Additional parameters for vector lines
    ppars: dict
        Additional parameters for point scatter

    :group: output.seq_plugins

    """

    def __init__(self, show_p=True, show_v=True, vpars={}, **ppars):
        """
        Constructor.

        Parameters
        ----------
        show_p: bool
            Flag for showing wake points
        show_v: bool
            Flag for showing wake vectors
        vpars: dict
            Additional parameters for vector lines
        ppars: dict, optional
            Additional parameters for point scatter

        """
        super().__init__()
        self.show_p = show_p
        self.show_v = show_v

        self.vpars = dict(color="blue")
        self.vpars.update(vpars)

        self.ppars = dict(color="blue")
        self.ppars.update(ppars)

    def initialize(self, algo):
        """
        Initialize data based on the intial iterator

        Parameters
        ----------
        algo: foxes.algorithms.sequential.Sequential
            The current sequential algorithm

        """
        super().initialize(algo)
        self._data = []

    def update(self, algo, fres, pres=None):
        """
        Updates data based on current iteration

        Parameters
        ----------
        algo: foxes.algorithms.sequential.Sequential
            The latest sequential algorithm
        fres: xarray.Dataset
            The latest farm results
        pres: xarray.Dataset, optional
            The latest point results

        """
        super().update(algo, fres, pres)

        wframe = algo.wake_frame
        if not isinstance(wframe, SeqDynamicWakes):
            raise ValueError(
                f"Wake frame not of type SeqDynamicWakes, got {type(algo.wake_frame).__name__}"
            )

        counter = algo.counter
        N = counter + 1
        dt = wframe._dt[counter] if counter < len(wframe._dt) else wframe._dt[-1]

        self._data.append(
            (
                dt,
                wframe._traces_p[:N].copy(),
                wframe._traces_v[:N].copy(),
            )
        )

    def gen_images(self, ax):
        """

        Parameters
        ----------
        ax: matplotlib.Axis
            The plotting axis

        Yields
        ------
        imgs: tuple
            The (figure, artists) tuple

        """
        while len(self._data):
            dt, pts, v = self._data.pop(0)

            N = len(pts)
            artists = []
            if self.show_p:
                artists += [
                    ax.scatter(
                        pts[:, downwind_index, 0],
                        pts[:, downwind_index, 1],
                        animated=True,
                        **self.ppars,
                    )
                    for downwind_index in range(self.algo.n_turbines)
                ]

            if self.show_v:
                for downwind_index in range(self.algo.n_turbines):
                    for i in range(N):
                        p = pts[i, downwind_index]
                        dxy = v[i, downwind_index] * dt
                        artists.append(
                            ax.arrow(
                                p[0],
                                p[1],
                                dxy[0],
                                dxy[1],
                                length_includes_head=True,
                                animated=True,
                                **self.vpars,
                            )
                        )

            yield ax.get_figure(), artists



import numpy as np
from foxes.core import PointDataModel
import foxes.variables as FV
import foxes.constants as FC


class Ustar2TI(PointDataModel):
    """
    Calculates TI from Ustar, using TI = Ustar / (kappa*WS)

    Attributes
    ----------
    max_ti: float
        Upper limit of the computed TI values

    :group: models.point_models

    """

    def __init__(self, max_ti=None, **kwargs):
        """
        Constructor

        Parameters
        ----------
        max_ti: float, optional
            Upper limit of the computed TI values
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(**kwargs)
        self.max_ti = max_ti

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return [FV.TI]

    def calculate(self, algo, mdata, fdata, pdata):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_points)

        """
        ustar = pdata[FV.USTAR]
        ws = pdata[FV.WS]

        ti = (ustar / FC.KAPPA) / ws
        if self.max_ti is not None:
            ti = np.maximum(ti, self.max_ti)

        return {FV.TI: ti}



import numpy as np
import argparse
import matplotlib.pyplot as plt

import foxes
import foxes.variables as FV


def calc(mbook, farm, states, wakes, points, args):
    algo = foxes.algorithms.Downwind(
        farm,
        states,
        wake_models=wakes,
        rotor_model=args.rotor,
        wake_frame="rotor_wd",
        partial_wakes=args.pwakes,
        mbook=mbook,
        verbosity=0,
    )

    farm_results = algo.calc_farm()
    point_results = algo.calc_points(farm_results, points[None, :])

    mbook.finalize(algo, verbosity=0)

    return point_results[args.var].to_numpy()


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-d",
        "--dists_D",
        help="turbine distances in units of D",
        type=float,
        nargs="+",
        default=[2.0, 5.0, 8.0, 12.0, 15.0, 20.0],
    )
    parser.add_argument(
        "-y",
        "--span_y",
        help="Span in y direction in units of D",
        type=float,
        default=4,
    )
    parser.add_argument(
        "-sp", "--step", help="Point step size in units of D", type=float, default=0.01
    )
    parser.add_argument(
        "-pc", "--plot_cols", help="Columns in the plot", type=int, default=3
    )
    parser.add_argument(
        "-hg", "--height", help="The point height", type=float, default=None
    )
    parser.add_argument("-v", "--var", help="The plot variable", default=FV.WS)
    parser.add_argument(
        "-fs", "--figsize_scale", help="Scale for single D plot", type=int, default=4
    )
    parser.add_argument("--ws", help="The wind speed", type=float, default=9.0)
    parser.add_argument("--wd", help="The wind direction", type=float, default=270.0)
    parser.add_argument("--ti", help="The TI value", type=float, default=0.05)
    parser.add_argument("--rho", help="The air density", type=float, default=1.225)
    parser.add_argument("--ct", help="Set CT by hand", default=None, type=float)
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=[
            "Bastankhah2014B_linear_k004",
            "Bastankhah2014_linear_k004",
            "Bastankhah2016_linear_k004",
        ],
        nargs="+",
    )
    parser.add_argument(
        "-ew",
        "--ewakes",
        help="The extra wake models",
        default=[],
        nargs="+",
    )
    parser.add_argument(
        "-m", "--tmodels", help="The turbine models", default=[], nargs="+"
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-p", "--pwakes", help="The partial wakes models", default="centre", nargs="+"
    )
    parser.add_argument(
        "-dfz",
        "--deficit",
        help="Plot the wind deficit instead of wind speed",
        action="store_true",
    )
    parser.add_argument("-e", "--engine", help="The engine", default="ProcessEngine")
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    args = parser.parse_args()

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype
    D = ttype.D
    H = ttype.H

    models = args.tmodels + [ttype.name]
    if args.ct is not None:
        mbook.turbine_models["set_ct"] = foxes.models.turbine_models.SetFarmVars()
        mbook.turbine_models["set_ct"].add_var(FV.CT, args.ct)
        models.append("set_ct")

    states = foxes.input.states.SingleStateStates(
        ws=args.ws, wd=args.wd, ti=args.ti, rho=args.rho
    )

    farm = foxes.WindFarm()
    farm.add_turbine(
        foxes.Turbine(
            xy=np.array([0.0, 0.0]),
            turbine_models=models,
        )
    )

    #  y lines:

    print("\nCalculating y lines\n")

    xlist = np.array(args.dists_D) * D
    ylist = np.arange(-args.span_y, args.span_y + args.step, args.step) * D
    nd = len(args.dists_D)
    nx = len(xlist)
    ny = len(ylist)
    points = np.zeros((nd, ny, 3))
    points[:, :, 2] = args.height if args.height is not None else H
    points[:, :, 0] = xlist[:, None]
    points[:, :, 1] = ylist[None, :]
    points = points.reshape(nd * ny, 3)

    ncols = min(args.plot_cols, nd)
    nrows = int(nd / ncols)
    while nrows * ncols < nd:
        nrows += 1

    figsize = (ncols * args.figsize_scale, nrows * args.figsize_scale)
    fig, axs = plt.subplots(nrows, ncols, figsize=figsize, sharex=True, sharey=True)
    with foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
    ):
        for wake in args.wakes:
            wakes = [wake] + args.ewakes
            print("Calculating:", wakes)

            results = calc(mbook, farm, states, wakes, points, args).reshape(nd, ny)

            for di, d in enumerate(args.dists_D):
                if nrows == 1 or ncols == 1:
                    ax = axs[di]
                else:
                    xi = int(di / ncols)
                    yi = di % ncols
                    ax = axs[xi, yi]

                if args.deficit:
                    dfz = (args.ws - results[di]) / args.ws
                    ax.plot(ylist / D, dfz, label=wake)
                    ax.set_ylabel("WS deficit")
                else:
                    ax.plot(ylist / D, results[di], label=wake)
                    ax.set_ylabel(args.var)

                ax.set_title(f"x = {d} D")
                ax.set_xlabel("y/D")
                ax.grid()

        ax.legend(loc="best")
        if not args.nofig:
            plt.show()
        plt.close(fig)

        # x line:

        print("\nCalculating x line\n")

        xlist = np.arange(-1, args.dists_D[-1] + args.step, args.step) * D
        nx = len(xlist)
        points = np.zeros((nx, 3))
        points[:, 2] = args.height if args.height is not None else H
        points[:, 0] = xlist

        figsize = (args.plot_cols * args.figsize_scale, args.figsize_scale)
        fig, ax = plt.subplots(figsize=figsize)

        for wake in args.wakes:
            wakes = [wake] + args.ewakes
            print("Calculating:", wakes)

            results = calc(mbook, farm, states, wakes, points, args)

            if args.deficit:
                dfz = (args.ws - results[0]) / args.ws
                ax.plot(xlist / D, dfz, label=wake)
                ax.set_ylabel("WS deficit")
            else:
                ax.plot(xlist / D, results[0], label=wake)
                ax.set_ylabel(args.var)

            ax.set_title("y = 0")
            ax.set_xlabel("x/D")
            ax.legend(loc="best")
            ax.grid()

    if not args.nofig:
        plt.show()



import pandas as pd
from pathlib import Path
import inspect

import foxes
import foxes.variables as FV


def test():
    thisdir = Path(inspect.getabsfile(inspect.currentframe())).parent
    print("TESTDIR:", thisdir)

    cfile = thisdir / "flappy" / "results.csv.gz"
    tPfile = thisdir / "NREL-5MW-D126-H90-P.csv"
    tCtfile = thisdir / "NREL-5MW-D126-H90-Ct.csv"
    sfile = thisdir / "states.csv.gz"
    lfile = thisdir / "test_farm.csv"

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFromTwo(
        data_source_P=tPfile,
        data_source_ct=tCtfile,
        col_ws_P_file="ws",
        col_ws_ct_file="ws",
        col_P="P",
        col_ct="ct",
        P_nominal=5000,
        H=90.0,
        D=126,
        name="power_and_ct_curve",
    )
    mbook.turbine_types[ttype.name] = ttype

    states = foxes.input.states.StatesTable(
        data_source=sfile,
        output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO],
        var2col={FV.WS: "ws", FV.WD: "wd", FV.TI: "ti"},
        fixed_vars={FV.RHO: 1.225},
    )

    farm = foxes.WindFarm()
    foxes.input.farm_layout.add_from_file(
        farm, lfile, turbine_models=[ttype.name], verbosity=0
    )

    algo = foxes.algorithms.Downwind(
        farm=farm,
        states=states,
        mbook=mbook,
        rotor_model="centre",
        wake_models=["Jensen_linear_k007"],
        wake_frame="rotor_wd",
        partial_wakes={"Jensen_linear_k007": "top_hat"},
        verbosity=0,
    )

    with foxes.Engine.new("threads", chunk_size_states=2000):
        data = algo.calc_farm()

    df = data.to_dataframe()[[FV.AMB_WD, FV.WD, FV.AMB_REWS, FV.REWS, FV.AMB_P, FV.P]]
    df = df.reset_index()

    print()
    print("TRESULTS\n")
    print(df)

    # print("\Reading file", cfile)
    fdata = pd.read_csv(cfile)
    print(fdata)

    print("\nVERIFYING\n")
    df[FV.WS] = df["REWS"]
    df[FV.AMB_WS] = df["AMB_REWS"]

    # neglecting ws < 5 and ws > 20
    sel_ws = (
        (fdata[FV.WS] > 5) & (fdata[FV.WS] < 20) & (df["REWS"] > 5) & (df["REWS"] < 20)
    )

    # calculating difference
    delta = df.reset_index() - fdata
    delta = delta[sel_ws]
    print(delta)
    print(delta.max())
    chk = delta[[FV.AMB_WS, FV.AMB_P, FV.WS, FV.P]].abs()
    sel = chk[FV.WS] >= 1e-5
    print(sel)
    print(df[sel & sel_ws])
    print(fdata[sel & sel_ws])
    print(chk.loc[sel & sel_ws])
    print(chk.max())

    assert (chk[FV.WS] < 1e-5).all()
    assert (chk[FV.P] < 1e-3).all()


if __name__ == "__main__":
    test()



import numpy as np

from foxes.core import States
from foxes.config import config
import foxes.variables as FV
import foxes.constants as FC


class ScanStates(States):
    """
    Scan over selected variables

    Parameters
    ----------
    scans: dict
        The scans, key: variable name,
        value: scan values

    :group: input.states

    """

    def __init__(self, scans, **kwargs):
        """
        Constructor.

        Parameters
        ----------
        scans: dict
            The scans, key: variable name,
            value: scan values
        kwargs: dict, optional
            Parameters for the base class

        """
        super().__init__(**kwargs)
        self.scans = {v: np.asarray(d) for v, d in scans.items()}

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        n_v = len(self.scans)
        shp = [len(v) for v in self.scans.values()]
        self._N = np.prod(shp)
        self._vars = list(self.scans.keys())

        data = np.zeros(shp + [n_v], dtype=config.dtype_double)
        for i, d in enumerate(self.scans.values()):
            s = [None] * n_v
            s[i] = np.s_[:]
            s = tuple(s)
            data[..., i] = d[s]
        data = data.reshape(self._N, n_v)

        self.VARS = self.var("vars")
        self.DATA = self.var("data")
        idata = super().load_data(algo, verbosity)
        idata["coords"][self.VARS] = self._vars
        idata["data_vars"][self.DATA] = ((FC.STATE, self.VARS), data)

        return idata

    def set_running(
        self,
        algo,
        data_stash,
        sel=None,
        isel=None,
        verbosity=0,
    ):
        """
        Sets this model status to running, and moves
        all large data to stash.

        The stashed data will be returned by the
        unset_running() function after running calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data_stash: dict, optional
            Large data stash, this function adds data here, if given.
            Key: model name. Value: dict, large model data
        sel: dict, optional
            The subset selection dictionary
        isel: dict, optional
            The index subset selection dictionary
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().set_running(algo, data_stash, sel, isel, verbosity)

        if data_stash is not None:
            data_stash[self.name].update(dict(scans=self.scans))
        del self.scans

    def unset_running(
        self,
        algo,
        data_stash,
        sel=None,
        isel=None,
        verbosity=0,
    ):
        """
        Sets this model status to not running, recovering large data
        from stash

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data_stash: dict, optional
            Reconstruct model data from this stash, if given.
            Key: model name. Value: dict, large model data
        sel: dict, optional
            The subset selection dictionary
        isel: dict, optional
            The index subset selection dictionary
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().unset_running(algo, data_stash, sel, isel, verbosity)

        if data_stash is not None:
            data = data_stash[self.name]
            self.scans = data.pop("scans")

    def size(self):
        """
        The total number of states.

        Returns
        -------
        int:
            The total number of states

        """
        return self._N

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self._vars

    def calculate(self, algo, mdata, fdata, tdata):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape
            (n_states, n_targets, n_tpoints)

        """
        self.ensure_output_vars(algo, tdata)

        for i, v in enumerate(self._vars):
            if v not in tdata:
                tdata[v] = np.zeros_like(tdata[FC.TARGETS][..., 0])
            tdata[v][:] = mdata[self.DATA][:, None, None, i]

        # add weights:
        tdata[FV.WEIGHT] = np.full(
            (mdata.n_states, 1, 1), 1 / self._N, dtype=config.dtype_double
        )
        tdata.dims[FV.WEIGHT] = (FC.STATE, FC.TARGET, FC.TPOINT)

        return {v: tdata[v] for v in self.output_point_vars(algo)}



import numpy as np

from foxes.core import Turbine
from foxes.utils import wd2wdvec


def add_ring(
    farm,
    xy_base,
    dist,
    n_turbines,
    offset_deg=0,
    indices=None,
    names=None,
    verbosity=1,
    **turbine_parameters,
):
    """
    Add a ring of turbines.

    Parameters
    ----------
    farm: foxes.WindFarm
        The wind farm
    xy_base: numpy.ndarray
        The base point, shape: (2,)
    dist: float
        The distance between turbines
    n_turbines: int
        The number of turbines
    offset_deg: float
        The offset from north in degrees,
        following wind direction conventions
    indices: list of int, optional
        The turbine indices
    names: list of str, optional
        The turbine names
    verbosity: int
        The verbosity level, 0 = silent
    turbine_parameters: dict, optional
        Parameters forwarded to `foxes.core.Turbine`

    :group: input.farm_layout

    """
    p0 = np.array(xy_base)
    R = n_turbines * dist / (2 * np.pi)
    a = np.atleast_1d(offset_deg)
    da = 360 / n_turbines

    for i in range(n_turbines):
        n = wd2wdvec(a)[0]

        farm.add_turbine(
            Turbine(
                xy=p0 + R * n,
                index=None if indices is None else indices[i],
                name=None if names is None else names[i],
                **turbine_parameters,
            ),
            verbosity=verbosity,
        )

        a[0] += da



from pathlib import Path
import inspect

import foxes
import foxes.variables as FV


def test():
    thisdir = Path(inspect.getabsfile(inspect.currentframe())).parent
    print("TESTDIR:", thisdir)

    ttype = "DTU10MW"
    sfile = "wind_rose_bremen.csv"
    lfile = thisdir / "test_farm.csv"
    cases = [
        (foxes.algorithms.Downwind, "rotor_wd"),
        (foxes.algorithms.Iterative, "rotor_wd"),
        (foxes.algorithms.Iterative, "rotor_wd_farmo"),
    ]
    lims = {FV.REWS: 5e-7, FV.P: 5e-4}

    base_results = None
    with foxes.Engine.new("threads", chunk_size_states=1000, n_procs=2):
        for Algo, frame in cases:
            print(f"\nENTERING CASE {(Algo.__name__, frame)}\n")

            mbook = foxes.models.ModelBook()

            states = foxes.input.states.StatesTable(
                data_source=sfile,
                output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO],
                var2col={FV.WS: "ws", FV.WD: "wd", FV.WEIGHT: "weight"},
                fixed_vars={FV.RHO: 1.225, FV.TI: 0.05},
            )

            farm = foxes.WindFarm()
            foxes.input.farm_layout.add_from_file(
                farm, lfile, turbine_models=[ttype], verbosity=1
            )

            kwargs = {}
            if Algo is foxes.algorithms.Iterative:
                kwargs["mod_cutin"] = {"modify_ct": False, "modify_P": False}

            algo = Algo(
                farm,
                states,
                mbook=mbook,
                rotor_model="grid16",
                wake_models=["Bastankhah2014_linear_k004", "IECTI2019_max"],
                wake_frame=frame,
                partial_wakes="rotor_points",
                verbosity=1,
                **kwargs,
            )

            # f Algo is foxes.algorithms.Iterative:
            #    algo.set_urelax("post_rotor", CT=0.9)

            data = algo.calc_farm()

            df = data.to_dataframe()[
                [FV.AMB_WD, FV.WD, FV.AMB_REWS, FV.REWS, FV.AMB_P, FV.P]
            ]

            print()
            print("TRESULTS\n")
            print(df)

            df = df.reset_index()

            if base_results is None:
                base_results = df

            else:
                print(f"CASE {(Algo.__name__, frame)}")
                delta = df - base_results
                print(delta)
                print(delta.min(), delta.max())

                for v, lim in lims.items():
                    chk = delta[v].abs().loc[df["AMB_REWS"] > 7]
                    print(f"CASE {(Algo.__name__, frame, v, lim)}:", chk.max())

                assert (chk < lim).all()


if __name__ == "__main__":
    test()



import numpy as np
import pandas as pd

from foxes.config import get_input_path

from .PCt_file import PCtFile


class TBLFile(PCtFile):
    """
    Reads turbine data from a TBL file.

    Examples
    --------
    A TBL file is a csv file with space as separator
    and two header lines. The followind lines denote
    wind speed, ct, P.

    - first row will be ignored
    - second row: H D ct-stand-still rated-power-in-MW
    - further rows: ws ct P-in-kW

    For example:

    45
    175.0 290.0 0.03 22.0
    3.0 0.820 102.9
    3.5 0.800 444.0
    4.0 0.810 874.7
    4.5 0.820 1418.7
    5.0 0.820 2100.9
    5.5 0.830 3021.2
    6.0 0.830 3904.7
    6.5 0.830 5061.7
    7.0 0.810 6379.0

    :group: models.turbine_types

    """

    def __init__(
        self,
        tbl_file,
        rho=1.225,
        **parameters,
    ):
        """
        Constructor.

        Parameters
        ----------
        tbl_file: str
            Path to the tbl file
        rho: float
            The air density for the curves
        paramerers: dict, optional
            Additional parameters for PCtFile class

        """
        fpath = get_input_path(tbl_file)
        assert fpath.suffix == ".tbl", f"Expecting *.tbl file, got '{tbl_file}'"

        meta = np.genfromtxt(fpath, skip_header=1, max_rows=1)
        sdata = pd.read_csv(
            fpath, sep=" ", skiprows=2, header=None, names=["ws", "ct", "P"]
        )

        super().__init__(
            sdata,
            col_ws="ws",
            col_P="P",
            col_ct="ct",
            H=meta[0],
            D=meta[1],
            P_nominal=meta[3] * 1e3,
            P_unit="kW",
            rho=rho,
            **parameters,
        )



import numpy as np
import pandas as pd

from foxes.utils import Dict
from foxes.core import Turbine, TurbineType, WindFarm
from foxes.models.farm_controllers import BasicFarmController, OpFlagController
import foxes.variables as FV
import foxes.constants as FC


def read_turbine_types(wio_farm, mbook, ws_exp_P, ws_exp_ct, verbosity):
    """
    Reads the turbine type from windio

    Parameters
    ----------
    wio_farm: dict
        The windio farm data
    mbook: foxes.models.ModelBook
        The model book
    ws_exp_P: int
        The REWS exponent for power
    ws_exp_ct: int
        The REWS exponent for ct
    verbosity: int
        The verbosity level, 0=silent

    Returns
    -------
    ttypes: dict
        Mapping from turbine type key to turbine
        type name in the model book

    :group: input.yaml.windio

    """

    def _print(*args, level=1, **kwargs):
        if verbosity >= level:
            print(*args, **kwargs)

    if "turbine_types" not in wio_farm:
        wio_farm["turbine_types"] = Dict(
            {0: wio_farm["turbines"]}, _name="turbine_types"
        )

    ttypes = {}
    for k, wio_trbns in wio_farm["turbine_types"].items():
        tname = wio_trbns.pop_item("name")
        ttypes[k] = tname
        _print("    Reading turbine type", k, level=3)
        _print("      Name:", tname, level=3)
        _print("      Contents:", [k for k in wio_trbns.keys()], level=3)

        # read performance:
        performance = wio_trbns["performance"]
        _print("        Reading performance", level=3)
        _print("          Contents:", [k for k in performance.keys()], level=3)

        # P, ct data:
        if "power_curve" in performance:
            power_curve = performance["power_curve"]
            _print("            Reading power_curve", level=3)
            _print("              Contents:", [k for k in power_curve.keys()], level=3)
            P = power_curve["power_values"]
            ws_P = power_curve["power_wind_speeds"]
            ct_curve = performance["Ct_curve"]
            data_P = pd.DataFrame(data={"ws": ws_P, "P": P})

            P_max = np.max(data_P["P"])
            for P_unit, P_scale in [("W", 1e6), ("kW", 1e3), ("MW", 1)]:
                if P_max / P_scale >= 1:
                    break
            assert P_max / P_scale >= 1 and P_max / P_scale < 1000, (
                f"Failed to determin P_unit for max power {P_max}"
            )
            _print(f"              Determined P_unit = {P_unit}", level=3)

            _print("            Reading Ct_curve", level=3)
            _print("              Contents:", [k for k in ct_curve.keys()], level=3)
            ct = ct_curve["Ct_values"]
            ws_ct = ct_curve["Ct_wind_speeds"]
            data_ct = pd.DataFrame(data={"ws": ws_ct, "ct": ct})

            def _get_wse_var(wse):
                if wse not in [1, 2, 3]:
                    raise ValueError(
                        f"Expecting wind speed exponent 1, 2 or 3, got {wse}"
                    )
                return FV.REWS if wse == 1 else (FV.REWS2 if wse == 2 else FV.REWS3)

            _print(f"            Creating model '{tname}'", level=3)
            _print("              Turbine type class: PCtFomTwo", level=3)
            mbook.turbine_types[tname] = TurbineType.new(
                ttype_type="PCtFromTwo",
                data_source_P=data_P,
                data_source_ct=data_ct,
                col_ws_P_file="ws",
                col_ws_ct_file="ws",
                col_P="P",
                col_ct="ct",
                H=wio_trbns["hub_height"],
                D=wio_trbns["rotor_diameter"],
                var_ws_ct=_get_wse_var(ws_exp_ct),
                var_ws_P=_get_wse_var(ws_exp_P),
                rho=1.225,
                P_unit=P_unit,
            )
            _print("               ", mbook.turbine_types[tname], level=3)

        # P, ct data:
        elif "Cp_curve" in performance:
            cp_curve = performance["Cp_curve"]
            _print("            Reading Cp_curve", level=3)
            _print("              Contents:", [k for k in cp_curve.keys()], level=3)
            cp = cp_curve["Cp_values"]
            ws_cp = cp_curve["Cp_wind_speeds"]
            ct_curve = performance["Ct_curve"]
            _print("            Reading Ct_curve", level=3)
            _print("              Contents:", [k for k in ct_curve.keys()], level=3)
            ct = ct_curve["Ct_values"]
            ws_ct = ct_curve["Ct_wind_speeds"]

            data_cp = pd.DataFrame(data={"ws": ws_cp, "cp": cp})
            data_ct = pd.DataFrame(data={"ws": ws_ct, "ct": ct})

            _print(f"            Creating model '{tname}'", level=3)
            _print("              Turbine type class: CpCtFromTwo", level=3)
            mbook.turbine_types[tname] = TurbineType.new(
                ttype_type="CpCtFromTwo",
                data_source_cp=data_cp,
                data_source_ct=data_ct,
                col_ws_cp_file="ws",
                col_ws_ct_file="ws",
                col_cp="cp",
                col_ct="ct",
                H=wio_trbns["hub_height"],
                D=wio_trbns["rotor_diameter"],
            )
            _print("               ", mbook.turbine_types[tname], level=3)

        else:
            raise KeyError("Expecting either 'power_curve' or 'Cp_curve'")

    return ttypes


def read_layout(lname, ldict, farm, ttypes, verbosity=1):
    """
    Read wind farm layout from windio input

    Parameters
    ----------
    lname: str
        The layout name
    ldict: dict
        The layout data
    farm: foxes.core.WindFarm
        The wind farm
    ttypes: dict
        Mapping from turbine type key to turbine
        type name in the model book
    verbosity: int
        The verbosity level, 0=silent

    :group: input.yaml.windio

    """
    if verbosity > 2:
        print(f"        Reading '{lname}'")
    cdict = ldict["coordinates"]
    tmap = ldict.get_item("turbine_types", None)
    if verbosity > 2:
        print("          Turbine type map:", tmap)
    for i, xy in enumerate(zip(cdict["x"], cdict["y"])):
        tt = ttypes[tmap[i] if tmap is not None else 0]
        farm.add_turbine(
            Turbine(xy=np.array(xy), turbine_models=[tt]),
            verbosity=verbosity - 3,
        )
    if verbosity > 2:
        print(f"          Added {farm.n_turbines} turbines")


def read_farm(wio_dict, mbook, verbosity):
    """
    Reads the wind farm information

    Parameters
    ----------
    wio_dict: foxes.utils.Dict
        The windio data
    verbosity: int
        The verbosity level, 0=silent

    Returns
    -------
    farm: foxes.core.WindFarm
        The wind farm

    :group: input.yaml.windio

    """
    wio_farm = wio_dict["wind_farm"]
    if verbosity > 1:
        print("Reading wind farm")
        print("  Name:", wio_farm.pop_item("name", None))
        print("  Contents:", [k for k in wio_farm.keys()])

    # find REWS exponents:
    try:
        rotor_averaging = wio_dict["attributes"]["analysis"]["rotor_averaging"]
        ws_exp_P = rotor_averaging["wind_speed_exponent_for_power"]
        ws_exp_ct = rotor_averaging["wind_speed_exponent_for_ct"]
    except KeyError:
        ws_exp_P = 1
        ws_exp_ct = 1

    # create farm controller:
    if FV.OPERATING in wio_farm:
        op_dims, operating = wio_farm.pop_item(FV.OPERATING)
        assert (
            len(op_dims) == 2
            and op_dims[1] == FC.TURBINE
            and op_dims[0] in [FC.STATE, FC.TIME]
        ), f"Expecting operating field to have dims (state, turbine), got {op_dims}"
        mbook.farm_controllers["farm_cntrl"] = OpFlagController(operating)
    else:
        mbook.farm_controllers["farm_cntrl"] = BasicFarmController()
    if verbosity > 1:
        print(
            f"  Farm controller type: {type(mbook.farm_controllers['farm_cntrl']).__name__}"
        )

    # read turbine type:
    ttypes = read_turbine_types(wio_farm, mbook, ws_exp_P, ws_exp_ct, verbosity)

    # read layouts and create wind farm:
    farm = WindFarm()
    wfarm = wio_farm["layouts"]
    if isinstance(wfarm, dict):
        if "coordinates" in wfarm:
            wfarm = {"0": wfarm}
        layouts = Dict(wfarm, _name=wio_farm.name + ".layouts")
    else:
        layouts = {str(i): lf for i, lf in enumerate(wfarm)}
        layouts = Dict(layouts, _name=wio_farm.name + ".layouts")
    if verbosity > 2:
        print("    Reading layouts")
        print("      Contents:", [k for k in layouts.keys()])
    for lname, ldict in layouts.items():
        read_layout(lname, ldict, farm, ttypes, verbosity)

    return farm


def read_n_turbines(wio_dict):
    """
    Reads the number of turbines from windio input

    Parameters
    ----------
    wio_dict: foxes.utils.Dict
        The windio data

    Returns
    -------
    n_turbines: int
        The number of turbines

    :group: input.yaml.windio

    """
    wio_farm = wio_dict["wind_farm"]
    wfarm = wio_farm["layouts"]
    if isinstance(wfarm, dict):
        if "coordinates" in wfarm:
            wfarm = {"0": wfarm}
        layouts = Dict(wfarm, _name=wio_farm.name + ".layouts")
    else:
        layouts = {str(i): lf for i, lf in enumerate(wfarm)}
        layouts = Dict(layouts, _name=wio_farm.name + ".layouts")
    n_turbines = 0
    for ldict in layouts.values():
        n_turbines += len(ldict["coordinates"]["x"])

    return n_turbines


def read_hub_heights(wio_dict):
    """
    Reads the hub heights from windio input

    Parameters
    ----------
    wio_dict: foxes.utils.Dict
        The windio data

    Returns
    -------
    hub_heights: list of float
        The hub heights of all turbines

    :group: input.yaml.windio

    """
    wio_farm = wio_dict["wind_farm"]
    if "turbine_types" not in wio_farm:
        wio_farm["turbine_types"] = Dict(
            {0: wio_farm["turbines"]}, _name="turbine_types"
        )

    hub_heights = [tt["hub_height"] for tt in wio_farm["turbine_types"].values()]

    return hub_heights


def read_rotor_diameters(wio_dict):
    """
    Reads the rotor diameters from windio input

    Parameters
    ----------
    wio_dict: foxes.utils.Dict
        The windio data

    Returns
    -------
    rotor_diameters: list of float
        The rotor diameters of all turbines

    :group: input.yaml.windio

    """
    wio_farm = wio_dict["wind_farm"]
    if "turbine_types" not in wio_farm:
        wio_farm["turbine_types"] = Dict(
            {0: wio_farm["turbines"]}, _name="turbine_types"
        )

    rotor_diameters = [
        tt["rotor_diameter"] for tt in wio_farm["turbine_types"].values()
    ]

    return rotor_diameters



import numpy as np
from scipy.spatial.distance import cdist

from foxes.core import WakeFrame, TData
from foxes.utils import wd2uv
from foxes.algorithms.iterative import Iterative
from foxes.config import config
import foxes.variables as FV
import foxes.constants as FC


class DynamicWakes(WakeFrame):
    """
    Dynamic wakes for any kind of timeseries states.

    Attributes
    ----------
    max_age: int
        The maximal number of wake steps
    cl_ipars: dict
        Interpolation parameters for centre line
        point interpolation
    dt_min: float
        The delta t value in minutes,
        if not from timeseries data

    :group: models.wake_frames

    """

    def __init__(
        self,
        max_length_km=20,
        max_age=None,
        max_age_mean_ws=5,
        cl_ipars={},
        dt_min=None,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        max_length_km: float
            The maximal wake length in km
        max_age: int, optional
            The maximal number of wake steps
        max_age_mean_ws: float
            The mean wind speed for the max_age calculation,
            if the latter is not given
        cl_ipars: dict
            Interpolation parameters for centre line
            point interpolation
        dt_min: float, optional
            The delta t value in minutes,
            if not from timeseries data
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(max_length_km=max_length_km, **kwargs)

        self.max_age = max_age
        self.cl_ipars = cl_ipars
        self.dt_min = dt_min
        self._mage_ws = max_age_mean_ws

    def __repr__(self):
        return f"{type(self).__name__}(dt_min={self.dt_min}, max_length_km={self.max_length_km}, max_age={self.max_age})"

    def initialize(self, algo, verbosity=0):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        if not isinstance(algo, Iterative):
            raise TypeError(
                f"Incompatible algorithm type {type(algo).__name__}, expecting {Iterative.__name__}"
            )
        super().initialize(algo, verbosity)

        # disable subset state selection in iterative algo:
        algo.conv_crit.disable_subsets()

        # get and check times:
        times = np.asarray(algo.states.index())
        if self.dt_min is None:
            if not np.issubdtype(times.dtype, np.datetime64):
                raise TypeError(
                    f"{self.name}: Expecting state index of type np.datetime64, found {times.dtype}"
                )
            elif len(times) == 1:
                raise KeyError(
                    f"{self.name}: Expecting 'dt_min' for single step timeseries"
                )
            self._dt = (
                (times[1:] - times[:-1])
                .astype("timedelta64[s]")
                .astype(config.dtype_int)
            )
        else:
            n = max(len(times) - 1, 1)
            self._dt = np.full(n, self.dt_min * 60, dtype="timedelta64[s]").astype(
                config.dtype_int
            )
        self._dt = np.append(self._dt, self._dt[-1, None], axis=0)

        # find max age if not given:
        if self.max_age is None:
            step = np.mean(self._mage_ws * self._dt)
            self.max_age = max(int(self.max_length_km * 1e3 / step), 1)
            if verbosity > 0:
                print(
                    f"{self.name}: Assumed mean step = {step} m, setting max_age = {self.max_age}"
                )

        self.DATA = self.var("data")
        self.UPDATE = self.var("update")

    def calc_order(self, algo, mdata, fdata):
        """
        Calculates the order of turbine evaluation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data

        Returns
        -------
        order: numpy.ndarray
            The turbine order, shape: (n_states, n_turbines)

        """
        order = np.zeros((fdata.n_states, fdata.n_turbines), dtype=config.dtype_int)
        order[:] = np.arange(fdata.n_turbines)[None, :]
        return order

    def _calc_wakes(self, algo, mdata, fdata, downwind_index):
        """Helper function that computes the dynamic wakes"""
        # prepare:
        n_states = mdata.n_states
        rxyh = fdata[FV.TXYH][:, downwind_index]
        istates = mdata.chunki_states
        itargets = mdata.chunki_points
        prev_t = itargets
        i0 = mdata.states_i0(counter=True)
        i1 = i0 + n_states
        dt = self._dt[i0:i1]
        tdi = {
            v: (FC.STATE, FC.TARGET, FC.TPOINT)
            for v in algo.states.output_point_vars(algo)
        }
        key = f"{self.DATA}_{downwind_index}"
        self.AGE = self.var("age")
        self.XYHL = self.var("xyhl")

        def ukey_fun(fr, to):
            """helper function to create update key"""
            return f"{self.UPDATE}_dw{downwind_index}_from_{fr}_to_{to}"

        # compute wakes that start within this chunk: x, y, z, length
        data = algo.get_from_chunk_store(
            name=key, mdata=mdata, prev_t=prev_t, error=False
        )
        if data is None:
            assert itargets == 0, (
                "DynamicWakes: Missing chunk_store data for (istates, itargets) = ({istates}, {itargets}) at ({istates}, {itargets-prev_t})"
            )
            data = np.full(
                (n_states, self.max_age, 4), np.nan, dtype=config.dtype_double
            )
            data[:, 0, :3] = rxyh
            data[:, 0, 3] = 0
            tdt = {
                v: np.zeros((n_states, 1, 1), dtype=config.dtype_double)
                for v in tdi.keys()
            }
            pts = data[:, 0, :3].copy()
            for age in range(self.max_age - 1):
                if age == n_states:
                    break
                elif age == 0:
                    hmdata = mdata
                    hfdata = fdata
                    htdata = TData.from_points(points=pts[:, None], data=tdt, dims=tdi)
                    hdt = dt[:, None]
                else:
                    s = np.s_[age:]
                    pts = pts[:-1]
                    hmdata = mdata.get_slice(FC.STATE, s)
                    hfdata = fdata.get_slice(FC.STATE, s)
                    htdt = {v: d[s] for v, d in tdt.items()}
                    htdata = TData.from_points(points=pts[:, None], data=htdt, dims=tdi)
                    hdt = dt[s, None]
                    del htdt, s

                res = algo.states.calculate(algo, hmdata, hfdata, htdata)
                del hmdata, hfdata, htdata

                uv = wd2uv(res[FV.WD], res[FV.WS])[:, 0, 0]
                dxy = uv * hdt
                pts[:, :2] += dxy
                s = np.s_[:-age] if age > 0 else np.s_[:]
                data[s, age + 1, :3] = pts
                data[s, age + 1, 3] = data[s, age, 3] + np.linalg.norm(dxy, axis=-1)

                if age < self.max_age - 2:
                    s = ~np.isnan(data[:, age + 1, 3])
                    if np.min(data[s, age + 1, 3]) >= self.max_length_km * 1e3:
                        break

                del res, uv, s, hdt, dxy
            del pts, tdt

            # store this chunk's results:
            algo.add_to_chunk_store(
                key,
                data,
                dims=(FC.STATE, self.AGE, self.XYHL),
                mdata=mdata,
                copy=False,
            )
            algo.block_convergence(mdata=mdata)

        # apply updates from future chunks:
        for (jstates, jtargets), cdict in algo.chunk_store.items():
            uname = ukey_fun(jstates, istates)
            if jstates > istates and jtargets == 0 and uname in cdict:
                u = cdict[uname]
                if u is not None:
                    sel = np.isnan(data) & ~np.isnan(u)
                    if np.any(sel):
                        data[:] = np.where(sel, u, data)
                        algo.block_convergence(mdata=mdata)
                    cdict[uname] = None
                    del sel
                del u

        # compute wakes from previous chunks:
        prev_s = 0
        wi0 = i0
        data = [data]
        while istates - prev_s > 0:
            prev_s += 1

            # read data from previous chunk:
            hdata, (h_i0, h_n_states, __, __) = algo.get_from_chunk_store(
                name=key,
                mdata=mdata,
                prev_s=prev_s,
                prev_t=prev_t,
                ret_inds=True,
                error=False,
            )
            if hdata is None:
                break
            else:
                hdata = hdata.copy()
                wi0 = h_i0

                # select points with index+age=i0:
                sts = np.arange(h_n_states)
                ags = i0 - (h_i0 + sts)
                sel = ags < self.max_age - 1
                if np.any(sel):
                    sts = sts[sel]
                    ags = ags[sel]
                    pts = hdata[sts, ags, :3]
                    sel = (
                        np.all(~np.isnan(pts[:, :2]), axis=-1)
                        & np.any(np.isnan(hdata[sts, ags + 1, :2]), axis=-1)
                        & (hdata[sts, ags, 3] <= self.max_length_km * 1e3)
                    )
                    if np.any(sel):
                        sts = sts[sel]
                        ags = ags[sel]
                        pts = pts[sel]
                        n_pts = len(pts)

                        tdt = {
                            v: np.zeros((n_states, n_pts, 1), dtype=config.dtype_double)
                            for v in algo.states.output_point_vars(algo)
                        }

                        # compute single state wake propagation:
                        isnan0 = np.isnan(hdata)
                        for si in range(n_states):
                            s = slice(si, si + 1, None)
                            hmdata = mdata.get_slice(FC.STATE, s)
                            hfdata = fdata.get_slice(FC.STATE, s)
                            htdt = {v: d[s] for v, d in tdt.items()}
                            htdata = TData.from_points(
                                points=pts[None, :], data=htdt, dims=tdi
                            )
                            hdt = dt[s, None]
                            del htdt, s

                            res = algo.states.calculate(algo, hmdata, hfdata, htdata)
                            del hmdata, hfdata, htdata

                            uv = wd2uv(res[FV.WD], res[FV.WS])[0, :, 0]
                            dxy = uv * hdt
                            pts[:, :2] += dxy
                            del res, uv, hdt

                            ags += 1
                            hdata[sts, ags, :3] = pts
                            hdata[sts, ags, 3] = hdata[
                                sts, ags - 1, 3
                            ] + np.linalg.norm(dxy, axis=-1)
                            del dxy

                            hsel = (h_i0 + sts + ags < i1) & (ags < self.max_age - 1)
                            if np.any(hsel):
                                sts = sts[hsel]
                                ags = ags[hsel]
                                pts = pts[hsel]
                                tdt = {v: d[:, hsel] for v, d in tdt.items()}
                                del hsel
                            else:
                                del hsel
                                break

                        # store update:
                        sel = isnan0 & (~np.isnan(hdata))
                        if np.any(sel):
                            udata = np.full_like(hdata, np.nan)
                            udata[sel] = hdata[sel]
                            algo.add_to_chunk_store(
                                ukey_fun(istates, istates - prev_s),
                                udata,
                                dims=(),
                                mdata=mdata,
                                copy=False,
                            )
                            algo.block_convergence(mdata=mdata)

                        del udata, tdt
                    del pts

                # store prev_s chunk's results:
                data.insert(0, hdata)

                del sts, ags, sel
            del hdata

        return np.concatenate(data, axis=0), wi0

    def get_wake_coos(
        self,
        algo,
        mdata,
        fdata,
        tdata,
        downwind_index,
    ):
        """
        Calculate wake coordinates of rotor points.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        tdata: foxes.core.TData
            The target point data
        downwind_index: int
            The index of the wake causing turbine
            in the downwind order

        Returns
        -------
        wake_coos: numpy.ndarray
            The wake frame coordinates of the evaluation
            points, shape: (n_states, n_targets, n_tpoints, 3)

        """
        # first compute dynamic wakes:
        wdata, wi0 = self._calc_wakes(algo, mdata, fdata, downwind_index)

        # prepare:
        targets = tdata[FC.TARGETS]
        n_states, n_targets, n_tpoints = targets.shape[:3]
        n_points = n_targets * n_tpoints
        points = targets.reshape(n_states, n_points, 3)
        rxyh = fdata[FV.TXYH][:, downwind_index]
        i0 = mdata.states_i0(counter=True)

        # initialize:
        wcoos = np.full((n_states, n_points, 3), 1e20, dtype=config.dtype_double)
        wcoos[:, :, 2] = points[:, :, 2] - rxyh[:, None, 2]
        wake_si = np.zeros((n_states, n_points), dtype=config.dtype_int)
        wake_si[:] = i0 + np.arange(n_states)[:, None]

        # find nearest wake point:
        for si in range(n_states):
            ags = np.arange(self.max_age)
            sts = i0 + si - ags - wi0
            sel = (sts >= 0) & (sts < len(wdata))
            if np.any(sel):
                sts = sts[sel]
                ags = ags[sel]
                sel = np.all(~np.isnan(wdata[sts, ags]), axis=-1)
                if np.any(sel):
                    sts = sts[sel]
                    ags = ags[sel]

                    dists = cdist(points[si, :, :2], wdata[sts, ags, :2])
                    j = np.argmin(dists, axis=1)
                    sts = sts[j]
                    ags = ags[j]
                    wake_si[si] = sts + wi0

                    nx = wdata[sts, ags, :2]
                    dp = points[si, :, :2] - nx
                    sel = ags < self.max_age - 1
                    if np.any(sel):
                        nx[sel] = wdata[sts[sel], ags[sel] + 1, :2] - nx[sel]
                    if np.any(~sel):
                        nx[~sel] -= wdata[sts[~sel], ags[~sel] - 1, :2]
                    dx = np.linalg.norm(nx, axis=-1)
                    nx /= dx[:, None] + 1e-14

                    projx = np.einsum("sd,sd->s", dp, nx)
                    sel = (projx > -dx) & (projx < dx)
                    if np.any(sel):
                        ny = np.concatenate([-nx[:, 1, None], nx[:, 0, None]], axis=1)

                        wcoos[si, sel, 0] = projx[sel] + wdata[sts[sel], ags[sel], 3]
                        wcoos[si, sel, 1] = np.einsum("sd,sd->s", dp[sel], ny[sel])

        # store turbines that cause wake:
        tdata[FC.STATE_SOURCE_ORDERI] = downwind_index

        # store states that cause wake for each target point,
        # will be used by model.get_data() during wake calculation:
        tdata.add(
            FC.STATES_SEL,
            wake_si.reshape(n_states, n_targets, n_tpoints),
            (FC.STATE, FC.TARGET, FC.TPOINT),
        )

        return algo.wake_deflection.calc_deflection(
            algo,
            mdata,
            fdata,
            tdata,
            downwind_index,
            wcoos.reshape(n_states, n_targets, n_tpoints, 3),
        )



"""
Wake superposition models.
"""

from .ws_linear import WSLinear as WSLinear
from .ws_linear import WSLinearLocal as WSLinearLocal

from .ws_pow import WSPow as WSPow
from .ws_pow import WSPowLocal as WSPowLocal

from .ws_max import WSMax as WSMax
from .ws_max import WSMaxLocal as WSMaxLocal

from .ws_quadratic import WSQuadratic as WSQuadratic
from .ws_quadratic import WSQuadraticLocal as WSQuadraticLocal

from .wind_vector import WindVectorLinear as WindVectorLinear
from .ws_product import WSProduct as WSProduct
from .ti_linear import TILinear as TILinear
from .ti_quadratic import TIQuadratic as TIQuadratic
from .ti_pow import TIPow as TIPow
from .ti_max import TIMax as TIMax



import numpy as np
import argparse
import matplotlib.pyplot as plt

import foxes
import foxes.variables as FV

if __name__ == "__main__":
    # define arguments and options:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-nt", "--n_t", help="The number of turbines", type=int, default=5
    )
    parser.add_argument(
        "-l",
        "--layout",
        help="The wind farm layout file (path or static)",
        default=None,
    )
    parser.add_argument("--ws", help="The wind speed", type=float, default=9.0)
    parser.add_argument("--wd", help="The wind direction", type=float, default=270.0)
    parser.add_argument("--ti", help="The TI value", type=float, default=0.08)
    parser.add_argument("--rho", help="The air density", type=float, default=1.225)
    parser.add_argument(
        "-dx", "--deltax", help="The turbine distance in x", type=float, default=800.0
    )
    parser.add_argument(
        "-dy", "--deltay", help="Turbine layout y step", type=float, default=0.0
    )
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument(
        "-m",
        "--tmodels",
        help="The turbine models",
        default=[],
        nargs="+",
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["Bastankhah2014_linear_k004", "CrespoHernandez_quadratic_ambka04"],
        nargs="+",
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-p", "--pwakes", help="The partial wakes models", default=None, nargs="+"
    )
    parser.add_argument(
        "-g", "--grounds", help="The ground models", default=None, nargs="+"
    )
    parser.add_argument("-f", "--frame", help="The wake frame", default="rotor_wd")
    parser.add_argument("-v", "--var", help="The plot variable", default=FV.WS)
    parser.add_argument("-e", "--engine", help="The engine", default=None)
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-it", "--iterative", help="Use iterative algorithm", action="store_true"
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    args = parser.parse_args()

    # create model book
    mbook = foxes.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype

    # create states
    states = foxes.input.states.SingleStateStates(
        ws=args.ws,
        wd=args.wd,
        ti=args.ti,
        rho=args.rho,
    )

    # create wind farm
    print("\nCreating wind farm")
    farm = foxes.WindFarm()
    if args.layout is None:
        foxes.input.farm_layout.add_row(
            farm=farm,
            xy_base=np.array([0.0, 0.0]),
            xy_step=np.array([args.deltax, args.deltay]),
            n_turbines=args.n_t,
            turbine_models=args.tmodels + [ttype.name],
        )
    else:
        foxes.input.farm_layout.add_from_file(
            farm, args.layout, turbine_models=args.tmodels + [ttype.name]
        )

    # create algorithm
    Algo = foxes.algorithms.Iterative if args.iterative else foxes.algorithms.Downwind
    algo = Algo(
        farm,
        states,
        rotor_model=args.rotor,
        wake_models=args.wakes,
        wake_frame=args.frame,
        partial_wakes=args.pwakes,
        ground_models=args.grounds,
        mbook=mbook,
        engine=args.engine,
        n_procs=args.n_cpus,
        chunk_size_points=args.chunksize_points,
        verbosity=1,
    )

    farm_results = algo.calc_farm()
    print("\nResults data:\n", farm_results)

    # add capacity and efficiency to farm results
    o = foxes.output.FarmResultsEval(farm_results)
    o.add_capacity(algo)
    o.add_capacity(algo, ambient=True)
    o.add_efficiency()

    # state-turbine results
    farm_df = farm_results.to_dataframe()
    print("\nFarm results data:\n")
    print(
        farm_df[
            [
                FV.X,
                FV.ORDER,
                FV.WD,
                FV.AMB_REWS,
                FV.REWS,
                FV.AMB_TI,
                FV.TI,
                FV.AMB_P,
                FV.P,
                FV.CT,
                FV.EFF,
            ]
        ]
    )
    print()

    # results by turbine
    turbine_results = o.reduce_states(
        {
            FV.AMB_P: "weights",
            FV.P: "weights",
            FV.AMB_CAP: "weights",
            FV.CAP: "weights",
        }
    )
    turbine_results[FV.AMB_YLD] = o.calc_turbine_yield(
        algo=algo, annual=True, ambient=True
    )
    turbine_results[FV.YLD] = o.calc_turbine_yield(algo=algo, annual=True)
    turbine_results[FV.EFF] = turbine_results[FV.P] / turbine_results[FV.AMB_P]
    print("\nResults by turbine:\n")
    print(turbine_results)

    # power results
    P0 = o.calc_mean_farm_power(ambient=True)
    P = o.calc_mean_farm_power()
    print(f"\nFarm power        : {P / 1000:.1f} MW")
    print(f"Farm ambient power: {P0 / 1000:.1f} MW")
    print(f"Farm efficiency   : {o.calc_farm_efficiency() * 100:.2f} %")
    print(f"Annual farm yield : {turbine_results[FV.YLD].sum():.2f} GWh.")

    if not args.nofig:
        # horizontal flow plot
        o = foxes.output.FlowPlots2D(algo, farm_results)
        g = o.gen_states_fig_xy(
            args.var, resolution=10, rotor_color="red", figsize=(10, 3)
        )
        fig = next(g)
        plt.show()
        plt.close(fig)

        # vertical flow plot
        o = foxes.output.FlowPlots2D(algo, farm_results)
        g = o.gen_states_fig_xz(
            args.var,
            resolution=10,
            x_direction=np.mod(args.wd, 360.0),
            rotor_color="red",
            figsize=(10, 3),
        )
        fig = next(g)
        plt.show()



import argparse

import foxes
import foxes.variables as FV

if __name__ == "__main__":
    # define arguments:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-P",
        "--P_percent",
        help="Power percent choice, applied to all states",
        type=float,
        default=55.0,
    )
    parser.add_argument(
        "-s",
        "--states",
        help="The states input file (path or static)",
        default="wind_rose_bremen.csv",
    )
    parser.add_argument(
        "-l",
        "--layout",
        help="The wind farm layout file (path or static)",
        default="test_farm_67.csv",
    )
    parser.add_argument(
        "-D",
        help="The rotor diameter",
        default=126.0,
        type=float,
    )
    parser.add_argument(
        "-H",
        help="The hub height",
        default=90.0,
        type=float,
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-p", "--pwakes", help="The partial wakes models", default=None, nargs="+"
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["Bastankhah2014_linear_lim_k004"],
        nargs="+",
    )
    parser.add_argument(
        "-m", "--tmodels", help="The turbine models", default=[], nargs="+"
    )
    parser.add_argument("-e", "--engine", help="The engine", default="default")
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    args = parser.parse_args()

    mbook = foxes.models.ModelBook()
    mbook.turbine_types["ttypeDH"] = foxes.models.turbine_types.NullType(
        D=args.D, H=args.H
    )

    mbook.turbine_models["lookup"] = foxes.models.turbine_models.LookupTable(
        "curtail_to_power.csv",
        input_vars=[FV.REWS, "P_percent"],
        output_vars=[FV.P, FV.CT],
        varmap={
            FV.REWS: "wind",
            "P_percent": "powerPercent",
            FV.P: "GenPower",
            FV.CT: "ct",
        },
        P_percent=args.P_percent,
    )

    states = foxes.input.states.StatesTable(
        data_source=args.states,
        output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO],
        var2col={FV.WS: "ws", FV.WD: "wd", FV.WEIGHT: "weight"},
        fixed_vars={FV.RHO: 1.225, FV.TI: 0.05},
    )

    farm = foxes.WindFarm()
    foxes.input.farm_layout.add_from_file(
        farm,
        args.layout,
        col_x="x",
        col_y="y",
        col_H="H",
        turbine_models=["ttypeDH", "lookup"] + args.tmodels,
    )

    algo = foxes.algorithms.Downwind(
        farm,
        states,
        rotor_model=args.rotor,
        wake_models=args.wakes,
        wake_frame="rotor_wd",
        partial_wakes=args.pwakes,
        mbook=mbook,
    )

    with foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
    ):
        farm_results = algo.calc_farm()

    fr = farm_results.to_dataframe()
    print(fr[[FV.WD, FV.H, FV.AMB_REWS, FV.REWS, FV.AMB_P, FV.P, FV.CT, FV.WEIGHT]])

    o = foxes.output.FarmResultsEval(farm_results)
    P0 = o.calc_mean_farm_power(ambient=True)
    P = o.calc_mean_farm_power()
    print(f"\nFarm power        : {P / 1000:.1f} MW")
    print(f"Farm ambient power: {P0 / 1000:.1f} MW")
    print(f"Farm efficiency   : {o.calc_farm_efficiency() * 100:.2f} %")
    print(f"Annual farm yield : {o.calc_farm_yield(algo=algo):.2f} GWh")



def all_subclasses(cls):
    """
    Searches all classes derived from some
    base class.

    Parameters
    ----------
    cls: class
        The base class

    Returns
    -------
    list of class:
        The derived classes

    :group: utils

    """
    return set(cls.__subclasses__()).union(
        [s for c in cls.__subclasses__() for s in all_subclasses(c)]
    )


def new_cls(base_cls, cls_name):
    """
    Run-time class selector.

    Parameters
    ----------
    base_cls: object
        The base class
    cls_name: string
        Name of the class

    Returns
    -------
    cls: object
        The derived class

    :group: utils

    """

    if cls_name is None:
        return None

    allc = all_subclasses(base_cls)
    found = cls_name in [scls.__name__ for scls in allc]

    if found:
        for scls in allc:
            if scls.__name__ == cls_name:
                return scls

    else:
        estr = "Class '{}' not found, available classes derived from '{}' are \n {}".format(
            cls_name, base_cls.__name__, sorted([i.__name__ for i in allc])
        )
        raise KeyError(estr)


def new_instance(base_cls, cls_name, *args, **kwargs):
    """
    Run-time factory.

    Parameters
    ----------
    base_cls: object
        The base class
    cls_name: string
        Name of the class
    args: tuple, optional
        Additional parameters for the constructor
    kwargs: dict, optional
        Additional parameters for the constructor

    Returns
    -------
    obj: object
        The instance of the derived class

    :group: utils

    """

    cls = new_cls(base_cls, cls_name)
    if cls is None:
        return None
    else:
        return cls(*args, **kwargs)



"""
Output tools and functions.
"""

from .output import Output as Output
from .farm_layout import FarmLayoutOutput as FarmLayoutOutput
from .farm_results_eval import FarmResultsEval as FarmResultsEval
from .rose_plot import RosePlotOutput as RosePlotOutput
from .rose_plot import StatesRosePlotOutput as StatesRosePlotOutput
from .rose_plot import WindRoseBinPlot as WindRoseBinPlot
from .results_writer import ResultsWriter as ResultsWriter
from .state_turbine_map import StateTurbineMap as StateTurbineMap
from .turbine_type_curves import TurbineTypeCurves as TurbineTypeCurves
from .animation import Animator as Animator
from .calc_points import PointCalculator as PointCalculator
from .slice_data import SliceData as SliceData
from .slices_data import SlicesData as SlicesData
from .rotor_point_plots import RotorPointPlot as RotorPointPlot
from .state_turbine_table import StateTurbineTable as StateTurbineTable
from .plt import plt as plt

from .flow_plots_2d import FlowPlots2D as FlowPlots2D
from .seq_plugins import SeqFlowAnimationPlugin as SeqFlowAnimationPlugin
from .seq_plugins import SeqWakeDebugPlugin as SeqWakeDebugPlugin

from . import grids as grids
from . import seq_plugins as seq_plugins



import numpy as np
from xarray import Dataset, open_dataset

from foxes.config import config, get_input_path
from foxes.core import TurbineModel, States, MData, FData, TData
import foxes.constants as FC
import foxes.variables as FV


class PopulationStates(States):
    """
    States extended by a population factor.

    For each original state, n_pop states are created.
    This is useful for parameter studies, where each
    inserted state corresponds to a different value of the
    associated variable.

    Attributes
    ----------
    states: foxes.core.States
        The original states
    n_pop: int
        The population size

    :group: core

    """

    def __init__(self, states, n_pop):
        """
        Constructor.

        Parameters
        ----------
        states: foxes.core.States
            The original states
        n_pop: int
            The population size

        """
        super().__init__()
        self.states = states
        self.n_pop = n_pop

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        self.STATE0 = self.var(FC.STATE + "0")
        self.SMAP = self.var("SMAP")

        idata = super().load_data(algo, verbosity)
        idata0 = algo.get_model_data(self.states)
        n_states0 = self.states.size()
        for cname, coord in idata0["coords"].items():
            if cname != FC.STATE:
                idata["coords"][cname] = coord
            else:
                idata["coords"][self.STATE0] = coord

        for dname, (dims0, data0) in idata0["data_vars"].items():
            hdims = tuple(
                [d if d != FC.STATE else self.STATE0 for d in np.atleast_1d(dims0)]
            )
            idata["data_vars"][dname] = (hdims, data0)
        if FV.WEIGHT not in idata["data_vars"]:
            idata["data_vars"][FV.WEIGHT] = (
                (self.STATE0,),
                np.full(n_states0, 1 / n_states0, dtype=config.dtype_double),
            )

        smap = np.zeros((self.n_pop, self.states.size()), dtype=np.int32)
        smap[:] = np.arange(self.states.size())[None, :]
        smap = smap.reshape(self.size())
        idata["data_vars"][self.SMAP] = ((FC.STATE,), smap)

        found = False
        for dname, (dims0, data0) in idata["data_vars"].items():
            if self.STATE0 in dims0:
                found = True
                break
        if not found:
            del idata["coords"][self.STATE0]

        return idata

    def initialize(self, algo, verbosity=0):
        """
        Initializes the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        """
        if not self.states.initialized:
            self.states.initialize(algo, verbosity)
        super().initialize(algo, verbosity)

    def size(self):
        """
        The total number of states.

        Returns
        -------
        int:
            The total number of states

        """
        return self.states.size() * self.n_pop

    def output_point_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.states.output_point_vars(algo)

    def calculate(self, algo, mdata, fdata, tdata):
        """ "
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.Data
            The model data
        fdata: foxes.core.Data
            The farm data
        tdata: foxes.core.Data
            The point data

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_points)

        """
        smap = mdata[self.SMAP]

        def _map(in_data, DClass):
            if in_data is None:
                return None

            hdata = {}
            hdims = {}
            for dname, data in in_data.items():
                dms = in_data.dims[dname]
                if dname == self.SMAP or dname == self.STATE0:
                    pass
                elif dms[0] == self.STATE0:
                    hdata[dname] = data[smap]
                    hdims[dname] = tuple([FC.STATE] + list(dms)[1:])
                elif self.STATE0 in dms:
                    raise ValueError(
                        f"States '{self.name}': Found states variable not at dimension 0 for mdata entry '{dname}': {dms}"
                    )
                else:
                    hdata[dname] = data
                    hdims[dname] = dms
            return DClass(hdata, hdims, name=in_data.name + "_pop")

        hmdata = _map(mdata, MData)
        hfdata = _map(fdata, FData)
        htdata = _map(tdata, TData)
        out = self.states.calculate(algo, hmdata, hfdata, htdata)
        del hmdata, hfdata

        assert FV.WEIGHT in htdata, (
            f"Missing '{FV.WEIGHT}' in tdata results from states '{self.states.name}'"
        )
        out[FV.WEIGHT] = np.zeros(
            (htdata.n_states, htdata.n_targets, htdata.n_tpoints),
            dtype=config.dtype_double,
        )
        out[FV.WEIGHT][:] = htdata[FV.WEIGHT]

        return out


class PopulationModel(TurbineModel):
    """
    This model manages parameter studies by introducing
    a population into the states

    Attributes
    ----------
    index_coord: str
        The name of the index coordinate, labeling individuals
        within the population.
    turbine_coord: str
        The name of the turbine coordinate
    var2ncvar: dict
        Mapping from variable names to NetCDF variable names
    variables: list of str
        The variables to be set. If None, all variables
        fields from the dataset are used

    :group: core

    """

    def __init__(
        self,
        data_source,
        index_coord="index",
        turbine_coord="turbine",
        var2ncvar={},
        variables=None,
        verbosity=1,
        **kwargs,
    ):
        """
        Constructor

        Parameters
        ----------
        data_source: xarray.Dataset or str
            The population data or path to NetCDF file.
        index_coord: str
            The name of the index coordinate, labeling individuals
            within the population.
        turbine_coord: str
            The name of the turbine coordinate
        var2ncvar: dict
            Mapping from variable names to NetCDF variable names
        variables: list of str, optional
            The variables to be set. If None, all variables
            fields from the dataset are used#
        verbosity: int
            The verbosity level, 0 = silent
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(**kwargs)
        self.index_coord = index_coord
        self.turbine_coord = turbine_coord
        self.var2ncvar = var2ncvar
        self.variables = variables

        # n_pop is needed very early, hence the file is loaded here
        if isinstance(data_source, Dataset):
            self._data = data_source
        else:
            fpath = get_input_path(data_source)
            if verbosity > 0:
                print(
                    f"PopulationModel '{self.name}': Loading population data from '{fpath}'"
                )
            self._data = open_dataset(fpath)
        self.__n_pop = self._data.sizes[self.index_coord]

        if variables is None:
            ncvar2var = {ncv: v for v, ncv in self.var2ncvar.items()}
            self.variables = [
                ncvar2var.get(ncv, ncv)
                for ncv in self._data.data_vars.keys()
                if self._data[ncv].dims == (self.index_coord, self.turbine_coord)
            ]
            if verbosity > 0:
                print(
                    f"PopulationModel '{self.name}': Detected variables {self.variables}"
                )
            assert len(self.variables) > 0, (
                f"PopulationModel '{self.name}': No variables found in population data with dims ({self.index_coord}, {self.turbine_coord})"
            )

    @property
    def n_pop(self):
        """
        The population size

        Returns
        -------
        pop_size: int
            The population size

        """
        return self.__n_pop

    def output_farm_vars(self, algo):
        """
        The variables which are being modified by the model.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm

        Returns
        -------
        output_vars: list of str
            The output variable names

        """
        return self.variables

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        assert isinstance(algo.states, PopulationStates), (
            f"Algorithm '{algo.name}': PopulationModel '{self.name}' requires PopulationStates, found '{type(algo.states).__name__}'"
        )
        algo.init_states()

        n_states0 = algo.states.states.size()
        n_vrs = len(self.variables)
        data = np.zeros(
            (self.n_pop, n_states0, algo.n_turbines, n_vrs), dtype=config.dtype_double
        )
        for i, v in enumerate(self.variables):
            c = self.var2ncvar.get(v, v)
            assert c in self._data.data_vars, (
                f"PopulationModel '{self.name}': Missing variable '{c}' in population data, found {list(self._data.data_vars.keys())}"
            )
            data[..., i] = self._data.data_vars[c].values[:, None, :]
        data = data.reshape(algo.states.size(), algo.n_turbines, n_vrs)

        self.DATA = self.var("DATA")
        self.VARS = self.var("VARS")
        idata = super().load_data(algo, verbosity)
        idata["coords"][self.VARS] = self.variables
        idata["data_vars"][self.DATA] = ((FC.STATE, FC.TURBINE, self.VARS), data)

        return idata

    def set_running(
        self,
        algo,
        data_stash,
        sel=None,
        isel=None,
        verbosity=0,
    ):
        """
        Sets this model status to running, and moves
        all large data to stash.

        The stashed data will be returned by the
        unset_running() function after running calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data_stash: dict
            Large data stash, this function adds data here.
            Key: model name. Value: dict, large model data
        sel: dict, optional
            The subset selection dictionary
        isel: dict, optional
            The index subset selection dictionary
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().set_running(algo, data_stash, sel, isel, verbosity)
        data_stash[self.name] = dict(data=self._data)
        del self._data

    def unset_running(
        self,
        algo,
        data_stash,
        sel=None,
        isel=None,
        verbosity=0,
    ):
        """
        Sets this model status to not running, recovering large data
        from stash

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data_stash: dict
            Large data stash, this function adds data here.
            Key: model name. Value: dict, large model data
        sel: dict, optional
            The subset selection dictionary
        isel: dict, optional
            The index subset selection dictionary
        verbosity: int
            The verbosity level, 0 = silent

        """
        super().unset_running(algo, data_stash, sel, isel, verbosity)
        data = data_stash[self.name]
        self._data = data.pop("data")

    def calculate(self, algo, mdata, fdata, st_sel):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        mdata: foxes.core.MData
            The model data
        fdata: foxes.core.FData
            The farm data
        st_sel: slice or numpy.ndarray of bool
            The state-turbine selection,
            for shape: (n_states, n_turbines)

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray with shape (n_states, n_turbines)

        """
        self.ensure_output_vars(algo, fdata)

        data = mdata[self.DATA][st_sel]
        for i, v in enumerate(self.variables):
            fdata[v][st_sel] = data[..., i]

        return {v: fdata[v] for v in self.variables}

    def farm2pop_results(self, algo, farm_results):
        """
        Convert farm results to population results

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        farm_results: xarray.Dataset
            The farm results

        Returns
        -------
        pop_results: xarray.Dataset
            The population farm results

        """
        assert isinstance(algo.states, PopulationStates), (
            f"Algorithm '{algo.name}': PopulationModel '{self.name}' requires PopulationStates, found '{type(algo.states).__name__}'"
        )

        n_states0 = algo.states.states.size()
        inds0 = algo.states.states.index()
        coords = {FC.STATE: inds0} if inds0 is not None else {}
        coords.update(
            {c: d.values for c, d in farm_results.coords.items() if c != FC.STATE}
        )

        data = {}
        for dname, d in farm_results.data_vars.items():
            if d.dims[0] == FC.STATE:
                data[dname] = (
                    (self.index_coord,) + d.dims,
                    d.values.reshape((self.n_pop, n_states0) + d.shape[1:]),
                )
            else:
                data[dname] = (d.dims, d.values)

        return Dataset(data, coords=coords)



from concurrent.futures import ThreadPoolExecutor, ProcessPoolExecutor

from .pool import PoolEngine


class ThreadsEngine(PoolEngine):
    """
    The threads engine for foxes calculations.

    :group: engines

    """

    def __init__(self, *args, **kwargs):
        """
        Constructor
        """
        super().__init__(*args, share_cstore=True, **kwargs)

    def _create_pool(self):
        """Creates the pool"""
        self._pool = ThreadPoolExecutor(max_workers=self.n_workers)

    def submit(self, f, *args, **kwargs):
        """
        Submits a job to worker, obtaining a future

        Parameters
        ----------
        f: Callable
            The function f(*args, **kwargs) to be
            submitted
        args: tuple, optional
            Arguments for the function
        kwargs: dict, optional
            Arguments for the function

        Returns
        -------
        future: object
            The future object

        """
        return self._pool.submit(f, *args, **kwargs)

    def await_result(self, future):
        """
        Waits for result from a future

        Parameters
        ----------
        future: object
            The future

        Returns
        -------
        result: object
            The calculation result

        """
        return future.result()

    def _shutdown_pool(self):
        """Shuts down the pool"""
        self._pool.shutdown()


class ProcessEngine(ThreadsEngine):
    """
    The processes engine for foxes calculations.

    :group: engines

    """

    def _create_pool(self):
        """Creates the pool"""
        self._pool = ProcessPoolExecutor(max_workers=self.n_workers)



import numpy as np

from foxes.core import Turbine
from foxes.config import get_input_path
from foxes.models.turbine_types import TBLFile


def add_from_wrf(
    farm,
    directory,
    mbook,
    txt_file="windturbines.txt",
    tbl_name_fun=lambda i: f"wind-turbine-{i}.tbl",
    rho=1.225,
    verbosity=1,
    **turbine_parameters,
):
    """
    Add turbines to wind farm from a folder with tbl files
    and a txt file containing lat, lon, tbl_index

    Parameters
    ----------
    farm: foxes.WindFarm
        The wind farm
    directory: str
        The directory containing the tbl files and the txt file
    mbook: foxes.ModelBook, optional
        The model book, only needed if tbl_dir is specified
    txt_file: str
        The txt file name
    tbl_name_fun: function
        A function that takes an integer index and returns the corresponding tbl file name
    rho: float
        The air density for the turbine types, if tbl_dir is given
    verbosity: int
        The verbosity level, 0 = silent
    turbine_parameters: dict, optional
        Additional parameters are forwarded to the WindFarm.add_turbine().

    Examples
    --------
    TXT file format: lon, lat, tbl_index

    >>>    57.230095 -1.974240 1
    >>>    57.223583 -2.012743 1
    >>>    57.228312 -2.002316 1
    >>>    57.233467 -1.989579 2
    >>>    ...

    :group: input.farm_layout

    """
    assert farm.data_is_lonlat, "Require `input_is_lonlat=True` in WindFarm constructor"
    if verbosity > 0:
        print("Reading directory", directory)
    directory = get_input_path(directory)
    txt_path = directory / txt_file
    if not txt_path.exists():
        raise FileNotFoundError(f"File {txt_path} not found")
    data = np.genfromtxt(txt_path)
    ttypes = {}
    for i in np.unique(data[:, 2]).astype(int):
        tbl_path = directory / tbl_name_fun(i)
        ttypes[i] = tbl_path.stem
        if not tbl_path.exists():
            raise FileNotFoundError(f"File {tbl_path} not found")
        if verbosity > 1:
            print(f"Creating turbine type: {ttypes[i]}")
        mbook.turbine_types[ttypes[i]] = TBLFile(tbl_path, rho=rho)

    tmodels = turbine_parameters.pop("turbine_models", [])
    for i, row in enumerate(data):
        ttype = ttypes[int(row[-1])]
        farm.add_turbine(
            Turbine(
                index=i,
                xy=np.flip(row[:2]),
                H=None,
                D=None,
                turbine_models=[ttype] + tmodels,
                **turbine_parameters,
            ),
            verbosity=verbosity,
        )
    farm.lock(verbosity=verbosity)



import numpy as np

from foxes.utils import weibull_weights
from foxes.config import config
import foxes.variables as FV
import foxes.constants as FC

from .dataset_states import DatasetStates


class FieldData(DatasetStates):
    """
    Heterogeneous ambient states on a regular
    horizontal grid in NetCDF format.

    Attributes
    ----------
    states_coord: str
        The states coordinate name in the data
    x_coord: str
        The x coordinate name in the data
    y_coord: str
        The y coordinate name in the data
    h_coord: str
        The height coordinate name in the data
    weight_ncvar: str
        Name of the weight data variable in the nc file(s)
    interpn_pars: dict, optional
        Additional parameters for scipy.interpolate.interpn
    bounds_extra_space: float or str
        The extra space, either float in m,
        or str for units of D, e.g. '2.5D'
    height_bounds: tuple, optional
        The (h_min, h_max) height bounds in m. Defaults to H +/- 0.5*D

    Examples
    --------
    Simplistic example of the NetCDF structure:

    >>>    Dimensions:  (state: 2, h: 2, y: 2, x: 2)
    >>>    Coordinates:
    >>>    * state    (state) int32 8B 0 1
    >>>    * h        (h) float32 8B 0.0 300.0
    >>>    * y        (y) float32 8B 0.0 2.5e+03
    >>>    * x        (x) float32 8B 0.0 2.5e+03
    >>>    Data variables:
    >>>        ws       (state, h, y, x) float32 64B ...
    >>>        wd       (state, h, y, x) float32 64B ...

    :group: input.states

    """

    def __init__(
        self,
        *args,
        states_coord="Time",
        x_coord="UTMX",
        y_coord="UTMY",
        h_coord="height",
        weight_ncvar=None,
        bounds_extra_space=1000,
        height_bounds=None,
        interpn_pars={},
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        args: tuple, optional
            Arguments for the base class
        states_coord: str
            The states coordinate name in the data
        x_coord: str
            The x coordinate name in the data
        y_coord: str
            The y coordinate name in the data
        h_coord: str, optional
            The height coordinate name in the data
        weight_ncvar: str, optional
            Name of the weight data variable in the nc file(s)
        bounds_extra_space: float or str, optional
            The extra space, either float in m,
            or str for units of D, e.g. '2.5D'
        height_bounds: tuple, optional
            The (h_min, h_max) height bounds in m. Defaults to H +/- 0.5*D
        interpn_pars: dict
            Parameters for scipy.interpolate.interpn
        kwargs: dict, optional
            Additional parameters for the base class

        """
        super().__init__(*args, **kwargs)

        self.states_coord = states_coord
        self.x_coord = x_coord
        self.y_coord = y_coord
        self.h_coord = h_coord
        self.weight_ncvar = weight_ncvar
        self.interpn_pars = interpn_pars
        self.bounds_extra_space = bounds_extra_space
        self.height_bounds = height_bounds

        assert FV.WEIGHT not in self.ovars, (
            f"States '{self.name}': Cannot have '{FV.WEIGHT}' as output variable, got {self.ovars}"
        )
        self.variables = [v for v in self.ovars if v not in self.fixed_vars]
        if weight_ncvar is not None:
            self.var2ncvar[FV.WEIGHT] = weight_ncvar
            self.variables.append(FV.WEIGHT)
        elif FV.WEIGHT in self.var2ncvar:
            raise KeyError(
                f"States '{self.name}': Cannot have '{FV.WEIGHT}' in var2ncvar, use weight_ncvar instead"
            )

        self._cmap = {
            FC.STATE: self.states_coord,
            FV.X: self.x_coord,
            FV.Y: self.y_coord,
        }
        if self.h_coord is not None:
            self._cmap[FV.H] = self.h_coord

    def load_data(self, algo, verbosity=0):
        """
        Load and/or create all model data that is subject to chunking.

        Such data should not be stored under self, for memory reasons. The
        data returned here will automatically be chunked and then provided
        as part of the mdata object during calculations.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        idata: dict
            The dict has exactly two entries: `data_vars`,
            a dict with entries `name_str -> (dim_tuple, data_ndarray)`;
            and `coords`, a dict with entries `dim_name_str -> dim_array`

        """
        return super().load_data(
            algo,
            cmap=self._cmap,
            variables=self.variables,
            bounds_extra_space=self.bounds_extra_space,
            height_bounds=self.height_bounds,
            verbosity=verbosity,
        )


class WeibullField(FieldData):
    """
    Weibull sectors at regular grid points

    Attributes
    ----------
    wd_coord: str
        The wind direction coordinate name
    ws_coord: str
        The wind speed coordinate name, if wind speed bin
        centres are in data, else None
    ws_bins: numpy.ndarray
        The wind speed bins, including
        lower and upper bounds, shape: (n_ws_bins+1,)

    :group: input.states

    """

    def __init__(
        self,
        *args,
        wd_coord,
        ws_coord=None,
        ws_bins=None,
        **kwargs,
    ):
        """
        Constructor.

        Parameters
        ----------
        args: tuple, optional
            Positional arguments for the base class
        wd_coord: str
            The wind direction coordinate name
        ws_coord: str, optional
            The wind speed coordinate name, if wind speed bin
            centres are in data
        ws_bins: list of float, optional
            The wind speed bins, including
            lower and upper bounds
        kwargs: dict, optional
            Keyword arguments for the base class

        """
        super().__init__(
            *args,
            states_coord=wd_coord,
            time_format=None,
            load_mode="preload",
            **kwargs,
        )
        self.wd_coord = wd_coord
        self.ws_coord = ws_coord
        self.ws_bins = None if ws_bins is None else np.sort(np.asarray(ws_bins))

        assert ws_coord is None or ws_bins is None, (
            f"States '{self.name}': Cannot have both ws_coord '{ws_coord}' and ws_bins {ws_bins}"
        )
        assert ws_coord is not None or ws_bins is not None, (
            f"States '{self.name}': Expecting either ws_coord or ws_bins"
        )
        if ws_coord is not None:
            self._cmap[FV.WS] = ws_coord

        if FV.WD not in self.ovars:
            raise ValueError(
                f"States '{self.name}': Expecting output variable '{FV.WD}', got {self.ovars}"
            )
        for v in [FV.WEIBULL_A, FV.WEIBULL_k, FV.WEIGHT]:
            if v in self.ovars:
                raise ValueError(
                    f"States '{self.name}': Cannot have '{v}' as output variable"
                )
            if v not in self.variables:
                self.variables.append(v)

        for v in [FV.WS, FV.WD]:
            if v in self.variables:
                self.variables.remove(v)

        self._n_wd = None
        self._n_ws = None

    def __repr__(self):
        return f"{type(self).__name__}(n_wd={self._n_wd}, n_ws={self._n_ws})"

    def _read_ds(self, ds, cmap, variables, verbosity=0):
        """
        Helper function for _get_data, extracts data from the original Dataset.

        Parameters
        ----------
        ds: xarray.Dataset
            The Dataset to read data from
        cmap: dict
            A mapping from foxes variable names to Dataset dimension names
        variables: list of str
            The variables to extract from the Dataset
        verbosity: int
            The verbosity level, 0 = silent

        Returns
        -------
        coords: dict
            keys: Foxes variable names, values: 1D coordinate value arrays
        data: dict
            The extracted data, keys are variable names,
            values are tuples (dims, data_array)
            where dims is a tuple of dimension names and
            data_array is a numpy.ndarray with the data values

        """
        # read data, using wd_coord as state coordinate
        coords, data0 = super()._read_ds(ds, cmap, variables, verbosity)
        wd = coords.pop(FC.STATE)

        # replace state by wd coordinate
        data0 = {
            v: (tuple({FC.STATE: FV.WD}.get(c, c) for c in dims), d)
            for v, (dims, d) in data0.items()
        }

        # check weights
        if FV.WEIGHT not in data0:
            raise KeyError(
                f"States '{self.name}': Missing weights variable '{FV.WEIGHT}' in data, found {sorted(list(data0.keys()))}"
            )
        else:
            dims = data0[FV.WEIGHT][0]
            if FV.WD not in dims:
                raise KeyError(
                    f"States '{self.name}': Expecting weights variable '{FV.WEIGHT}' to contain dimension '{FV.WD}', got {dims}"
                )
            if FV.WS in dims:
                raise KeyError(
                    f"States '{self.name}': Expecting weights variable '{FV.WEIGHT}' to not contain dimension '{FV.WS}', got {dims}"
                )

        # construct wind speed bins and bin deltas
        assert FV.WS not in data0, (
            f"States '{self.name}': Cannot have '{FV.WS}' in data, found variables {list(data0.keys())}"
        )
        if self.ws_bins is not None:
            wsb = self.ws_bins
            wss = 0.5 * (wsb[:-1] + wsb[1:])
        elif self.ws_coord in ds.coords:
            wss = ds[self.ws_coord].to_numpy()
            wsb = np.zeros((len(wss) + 1,), dtype=config.dtype_double)
            wsb[1:-1] = 0.5 * (wss[1:] + wss[:-1])
            wsb[0] = wss[0] - 0.5 * wsb[1]
            wsb[-1] = wss[-1] + 0.5 * wsb[-2]
            self.ws_bins = wsb
        else:
            raise ValueError(
                f"States '{self.name}': Expecting ws_bins argument, or '{self.ws_coord}' among data coordinates, got {list(ds.coords.keys())}"
            )
        wsd = wsb[1:] - wsb[:-1]
        n_ws = len(wss)
        n_wd = len(wd)
        del wsb, ds

        # calculate Weibull weights
        dms = [FV.WS, FV.WD]
        shp = [n_ws, n_wd]
        for c in [FV.X, FV.Y, FV.H]:
            for v in [FV.WEIBULL_A, FV.WEIBULL_k]:
                if c in data0[v][0]:
                    dms.append(c)
                    shp.append(data0[v][1].shape[data0[v][0].index(c)])
                    break
        dms = tuple(dms)
        shp = tuple(shp)
        if data0[FV.WEIGHT][0] == dms:
            w = data0.pop(FV.WEIGHT)[1]
        else:
            s_w = tuple([np.s_[:] if c in data0[FV.WEIGHT][0] else None for c in dms])
            w = np.zeros(shp, dtype=config.dtype_double)
            w[:] = data0.pop(FV.WEIGHT)[1][s_w]
        s_ws = tuple([np.s_[:], None] + [None] * (len(dms) - 2))
        s_A = tuple([np.s_[:] if c in data0[FV.WEIBULL_A][0] else None for c in dms])
        s_k = tuple([np.s_[:] if c in data0[FV.WEIBULL_A][0] else None for c in dms])
        data0[FV.WEIGHT] = (
            dms,
            w
            * weibull_weights(
                ws=wss[s_ws],
                ws_deltas=wsd[s_ws],
                A=data0.pop(FV.WEIBULL_A)[1][s_A],
                k=data0.pop(FV.WEIBULL_k)[1][s_k],
            ),
        )
        del w, s_ws, s_A, s_k

        # translate binned data to states
        self._N = n_ws * n_wd
        self._inds = None
        data = {
            FV.WS: np.zeros((n_ws, n_wd), dtype=config.dtype_double),
            FV.WD: np.zeros((n_ws, n_wd), dtype=config.dtype_double),
        }
        data[FV.WS][:] = wss[:, None]
        data[FV.WD][:] = wd[None, :]
        data[FV.WS] = ((FC.STATE,), data[FV.WS].reshape(self._N))
        data[FV.WD] = ((FC.STATE,), data[FV.WD].reshape(self._N))
        for v in list(data0.keys()):
            dims, d = data0.pop(v)
            if dims[0] == FV.WD:
                dms = tuple([FC.STATE] + list(dims[1:]))
                shp = [n_ws] + list(d.shape)
                data[v] = np.zeros(shp, dtype=config.dtype_double)
                data[v][:] = d[None, ...]
                data[v] = (dms, data[v].reshape([self._N] + shp[2:]))
            elif len(dims) >= 2 and dims[:2] == (FV.WS, FV.WD):
                dms = tuple([FC.STATE] + list(dims[2:]))
                shp = [self._N] + list(d.shape[2:])
                data[v] = (dms, d.reshape(shp))
            else:
                data[v] = (dims, d)
        data0 = data

        return coords, data



import numpy as np
import pandas as pd
from pathlib import Path
import inspect

import foxes
import foxes.variables as FV


def test():
    thisdir = Path(inspect.getabsfile(inspect.currentframe())).parent
    print("TESTDIR:", thisdir)

    n_s = 30
    n_t = 52
    wd = 270.0
    ti = 0.08
    rotor = "centre"
    c = 100
    p0 = np.array([0.0, 0.0])
    stp = np.array([601.0, 15.0])
    cfile = thisdir / "flappy" / "results.csv.gz"
    tfile = thisdir / "NREL-5MW-D126-H90.csv"

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(
        data_source=tfile, var_ws_ct=FV.REWS, var_ws_P=FV.REWS
    )
    mbook.turbine_types[ttype.name] = ttype

    states = foxes.input.states.ScanStates(
        {
            FV.WS: np.linspace(6.0, 16.0, n_s),
            FV.WD: [wd],
            FV.TI: [ti],
            FV.RHO: [1.225],
        }
    )

    farm = foxes.WindFarm()
    foxes.input.farm_layout.add_row(
        farm=farm,
        xy_base=p0,
        xy_step=stp,
        n_turbines=n_t,
        turbine_models=[ttype.name],
        verbosity=1,
    )

    with foxes.Engine.new("threads", chunk_size_states=c):
        algo = foxes.algorithms.Downwind(
            farm,
            states,
            rotor_model=rotor,
            mbook=mbook,
            wake_models=["Jensen_linear_k007", "IECTI2005_max"],
            wake_frame="rotor_wd",
            partial_wakes=["top_hat", "top_hat"],
            verbosity=1,
        )

        data = algo.calc_farm()

        df = data.to_dataframe()[
            [
                FV.X,
                FV.Y,
                FV.WD,
                FV.AMB_REWS,
                FV.REWS,
                FV.AMB_TI,
                FV.TI,
                FV.AMB_CT,
                FV.CT,
            ]
        ]

        print("\nReading file", cfile)
        fdata = pd.read_csv(cfile).set_index(["state", "turbine"])

        print()
        print("TRESULTS\n")
        """sel = (df[FV.P] > 0) & (fdata[FV.P] > 0)
        df = df.loc[sel]
        fdata = fdata.loc[sel]"""
        print(df)
        print(fdata)

        print("\nVERIFYING\n")
        df[FV.WS] = df["REWS"]
        df[FV.AMB_WS] = df["AMB_REWS"]

        delta = df - fdata
        print(delta)

        chk = delta.abs()
        print(chk.max())

        var = FV.WS
        print(f"\nCHECKING {var}")
        sel = chk[var] >= 3e-3
        print(df.loc[sel])
        print(fdata.loc[sel])
        print(chk.loc[sel])
        assert (chk[var] < 3e-3).all()

        var = FV.TI
        print(f"\nCHECKING {var}")
        sel = chk[var] >= 3e-4
        print(df.loc[sel])
        print(fdata.loc[sel])
        print(chk.loc[sel])
        assert (chk[var] < 3e-4).all()

        var = FV.CT
        print(f"\nCHECKING {var}")
        sel = chk[var] >= 3e-5
        print(df.loc[sel])
        print(fdata.loc[sel])
        print(chk.loc[sel])
        assert (chk[var] < 3e-5).all()


if __name__ == "__main__":
    test()



from matplotlib import pyplot

from .output import Output


class plt(Output):
    """
    Class that runs plt commands

    :group: output

    """

    def __getattr__(self, name):
        return getattr(pyplot, name)

    def savefig(self, fname, *args, **kwargs):
        fpath = super().get_fpath(fname)
        pyplot.savefig(fpath, *args, **kwargs)



"""
Farm Optimization and eXtended yield Evaluation Software

"""

from .config import config as config
from .config import get_path as get_path
from .core import Engine as Engine
from .core import WindFarm as WindFarm
from .core import Turbine as Turbine
from .core import get_engine as get_engine
from .core import reset_engine as reset_engine
from .models import ModelBook as ModelBook

from .data import parse_Pct_file_name as parse_Pct_file_name
from .data import parse_Pct_two_files as parse_Pct_two_files
from .data import FARM as FARM
from .data import STATES as STATES
from .data import PCTCURVE as PCTCURVE
from .data import StaticData as StaticData

from . import algorithms as algorithms
from . import engines as engines
from . import models as models
from . import input as input
from . import output as output
from . import utils as utils

import importlib
from pathlib import Path

try:
    tomllib = importlib.import_module("tomllib")
    source_location = Path(__file__).parent
    if (source_location.parent / "pyproject.toml").exists():
        with open(source_location.parent / "pyproject.toml", "rb") as f:
            __version__ = tomllib.load(f)["project"]["version"]
    else:
        __version__ = importlib.metadata.version(__package__ or __name__)
except ModuleNotFoundError:
    __version__ = importlib.metadata.version(__package__ or __name__)



import numpy as np
import time
import argparse
from pathlib import Path

import flappy as fl
from flappy.config.variables import variables as FV

if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-c", "--chunksize", help="The maximal chunk size", type=int, default=500
    )
    parser.add_argument(
        "-o", "--ofile", help="The output file name", default="results.csv.gz"
    )
    parser.add_argument(
        "--n_cpus", help="The number of processors", type=int, default=4
    )
    args = parser.parse_args()

    c = args.chunksize
    p0 = np.array([0.0, 0.0])
    stp = np.array([500.0, 0.0])
    ofile = Path(args.ofile)
    tfile = "../NREL-5MW-D126-H90.csv"
    sfile = "../states.csv.gz"
    lfile = "../test_farm.csv"

    # init flappy:
    fl.init_flappy(n_cpus=args.n_cpus)

    # load model book:
    mbook = fl.ModelBook(ct_power_curve_file=tfile)

    # create wind farm:
    farm = fl.WindFarm()
    fl.input.add_turbines_from_csv(
        farm,
        lfile,
        col_index="index",
        col_x="x",
        col_y="y",
        col_h="H",
        rotor_diameter=126.0,
        rotor_model="centre",
        wake_models=["Jensen007_rotor"],
        turbine_models=["ct_P_curves"],
    )

    # create states:
    states = fl.input.AFSStatesTable(
        data_file=sfile,
        col_wd="wd",
        col_ws_ref="ws",
        col_ti="ti",
        col_mol="mol",
        z0=0.05,
        h_ref=100.0,
        col_weight="weight",
        air_density=1.225,
        max_chunk_size=args.chunksize,
    )
    states.initialize()

    time0 = time.time()

    # run calculation:
    results = farm.calculate(mbook, states, wake_superp=["wind_linear"])

    time1 = time.time()
    print("\nCalc time =", time1 - time0, "\n")

    df = results.state_turbine_results[[FV.WD, FV.AMB_WS, FV.WS, FV.AMB_P, FV.P]]

    print()
    print("TRESULTS\n")
    print(df)

    print("\nWriting file", ofile)
    df.to_csv(ofile)

    # close flappy:
    fl.shutdown_flappy()



"""
Vertical profile models.
"""

from .uniform import UniformProfile as UniformProfile
from .abl_log_neutral_ws import ABLLogNeutralWsProfile as ABLLogNeutralWsProfile
from .abl_log_stable_ws import ABLLogStableWsProfile as ABLLogStableWsProfile
from .abl_log_unstable_ws import ABLLogUnstableWsProfile as ABLLogUnstableWsProfile
from .abl_log_ws import ABLLogWsProfile as ABLLogWsProfile
from .sheared_ws import ShearedProfile as ShearedProfile
from .data_profile import DataProfile as DataProfile



import os
import numpy as np
from abc import ABC, abstractmethod
from tqdm import tqdm
from xarray import Dataset

from foxes.config import config, get_output_path
from foxes.utils import new_instance
from foxes.utils import write_nc as write_nc_file
import foxes.constants as FC

from .data import MData, FData, TData

__global_engine_data__ = dict(
    engine=None,
)


class Engine(ABC):
    """
    Abstract base clas for foxes calculation engines

    Attributes
    ----------
    chunk_size_states: int
        The size of a states chunk
    chunk_size_points: int
        The size of a points chunk
    progress_bar: bool, optional
        Use a progress bar instead of simply
        printing lines of reached percentages.
        Unless progress_bar is None, then neither
    verbosity: int
        The verbosity level, 0 = silent

    :group: core

    """

    def __init__(
        self,
        chunk_size_states=None,
        chunk_size_points=None,
        n_procs=None,
        progress_bar=True,
        verbosity=1,
    ):
        """
        Constructor.

        Parameters
        ----------
        chunk_size_states: int, optional
            The size of a states chunk
        chunk_size_points: int, optional
            The size of a points chunk
        n_procs: int, optional
            The number of processes to be used,
            or None for automatic
        progress_bar: bool, optional
            Use a progress bar instead of simply
            printing lines of reached percentages.
            Unless progress_bar is None, then neither
        verbosity: int
            The verbosity level, 0 = silent

        """
        self.chunk_size_states = chunk_size_states
        self.chunk_size_points = chunk_size_points
        self.progress_bar = progress_bar
        self.verbosity = verbosity

        try:
            self._n_procs = n_procs if n_procs is not None else os.process_cpu_count()
        except AttributeError:
            self._n_procs = os.cpu_count()
        self._n_workers = max(self._n_procs - 1, 1)

        self.__name = type(self).__name__
        self.__initialized = False
        self.__entered = False

    def __repr__(self):
        s = f"n_procs={self.n_procs}, chunk_size_states={self.chunk_size_states}, chunk_size_points={self.chunk_size_points}"
        return f"{self.name}({s})"

    def __enter__(self):
        if self.__entered:
            raise ValueError("Enter called for already entered engine")
        self.__entered = True
        if not self.initialized:
            self.initialize()
        return self

    def __exit__(self, *exit_args):
        if not self.__entered:
            raise ValueError("Exit called for not entered engine")
        self.__entered = False
        if self.initialized:
            self.finalize(*exit_args)

    def __del__(self):
        if self.initialized:
            self.finalize()

    @property
    def name(self):
        """
        The engine's name

        Returns
        -------
        nme: str
            The engine's name

        """
        return self.__name

    @property
    def n_procs(self):
        """
        The number of processes

        Returns
        -------
        n_procs: int
            The number of processes

        """
        return self._n_procs

    @property
    def n_workers(self):
        """
        The number of worker processes

        Returns
        -------
        n_workers: int
            The number of worker processes

        """
        return self._n_workers

    @property
    def has_progress_bar(self):
        """
        Flag for active progress bar

        Returns
        -------
        has_pbar: bool
            True if progress bar is active

        """
        return self.progress_bar is not None and self.progress_bar

    @property
    def prints_progress(self):
        """
        Flag for active progress printing

        Returns
        -------
        has_pbar: bool
            True if progress printing is active

        """
        return self.progress_bar is not None and not self.progress_bar

    @property
    def entered(self):
        """
        Flag that this model has been entered.

        Returns
        -------
        flag: bool
            True if the model has been entered.

        """
        return self.__entered

    @property
    def initialized(self):
        """
        Initialization flag.

        Returns
        -------
        flag: bool
            True if the model has been initialized.

        """
        return self.__initialized

    def initialize(self):
        """
        Initializes the engine.
        """
        if not self.entered:
            self.__enter__()
        elif not self.initialized:
            if get_engine(error=False, default=False) is not None:
                raise ValueError(
                    f"Cannot initialize engine '{self.name}', since engine already set to '{type(get_engine()).__name__}'"
                )
            __global_engine_data__["engine"] = self
            self.__initialized = True

    def finalize(self, type=None, value=None, traceback=None):
        """
        Finalizes the engine.

        Parameters
        ----------
        type: object, optional
            Dummy argument for the exit function
        value: object, optional
            Dummy argument for the exit function
        traceback: object, optional
            Dummy argument for the exit function

        """
        if self.entered:
            self.__exit__(type, value, traceback)
        elif self.initialized:
            __global_engine_data__["engine"] = None
            self.__initialized = False

    def print(self, *args, level=1, **kwargs):
        """Prints based on verbosity"""
        if self.verbosity >= level:
            print(*args, **kwargs)

    @abstractmethod
    def submit(self, f, *args, **kwargs):
        """
        Submits a job to worker, obtaining a future

        Parameters
        ----------
        f: Callable
            The function f(*args, **kwargs) to be
            submitted
        args: tuple, optional
            Arguments for the function
        kwargs: dict, optional
            Arguments for the function

        Returns
        -------
        future: object
            The future object

        """
        pass

    @abstractmethod
    def await_result(self, future):
        """
        Waits for result from a future

        Parameters
        ----------
        future: object
            The future

        Returns
        -------
        result: object
            The calculation result

        """
        pass

    @abstractmethod
    def map(
        self,
        func,
        inputs,
        *args,
        **kwargs,
    ):
        """
        Runs a function on a list of files

        Parameters
        ----------
        func: Callable
            Function to be called on each file,
            func(input, *args, **kwargs) -> data
        inputs: array-like
            The input data list
        args: tuple, optional
            Arguments for func
        kwargs: dict, optional
            Keyword arguments for func

        Returns
        -------
        results: list
            The list of results

        """
        pass

    @property
    def loop_dims(self):
        """
        Gets the loop dimensions (possibly chunked)

        Returns
        -------
        dims: list of str
            The loop dimensions (possibly chunked)

        """
        if self.chunk_size_states is None and self.chunk_size_states is None:
            return []
        elif self.chunk_size_states is None:
            return [FC.TARGET]
        elif self.chunk_size_points is None:
            return [FC.STATE]
        else:
            return [FC.STATE, FC.TARGET]

    def select_subsets(self, *datasets, sel=None, isel=None):
        """
        Takes subsets of datasets

        Parameters
        ----------
        datasets: tuple
            The xarray.Dataset or xarray.Dataarray objects
        sel: dict, optional
            The selection dictionary
        isel: dict, optional
            The index selection dictionary

        Returns
        -------
        subsets: list
            The subsets of the input data

        """
        if sel is not None:
            new_datasets = []
            for data in datasets:
                if data is not None:
                    s = {c: u for c, u in sel.items() if c in data.coords}
                    new_datasets.append(data.sel(s) if len(s) else data)
                else:
                    new_datasets.append(data)
            datasets = new_datasets

        if isel is not None:
            new_datasets = []
            for data in datasets:
                if data is not None:
                    s = {c: u for c, u in isel.items() if c in data.coords}
                    new_datasets.append(data.isel(s) if len(s) else data)
                else:
                    new_datasets.append(data)
            datasets = new_datasets

        return datasets

    def calc_chunk_sizes(self, n_states, n_targets=1):
        """
        Computes the sizes of states and points chunks

        Parameters
        ----------
        n_states: int
            The number of states
        n_targets: int
            The number of point targets

        Returns
        -------
        chunk_sizes_states: numpy.ndarray
            The sizes of all states chunks, shape: (n_chunks_states,)
        chunk_sizes_targets: numpy.ndarray
            The sizes of all targets chunks, shape: (n_chunks_targets,)

        """
        # determine states chunks:
        if self.chunk_size_states is None:
            n_chunks_states = min(self.n_workers, n_states)
            chunk_size_states = max(int(n_states / self.n_workers), 1)
        else:
            chunk_size_states = min(n_states, self.chunk_size_states)
            n_chunks_states = max(int(n_states / chunk_size_states), 1)
            if int(n_states / n_chunks_states) > chunk_size_states:
                n_chunks_states += 1
                chunk_size_states = int(n_states / n_chunks_states)

        # determine points chunks:
        chunk_sizes_targets = [n_targets]
        if n_targets > 1:
            if self.chunk_size_points is None:
                if n_targets < max(n_states, 1000):
                    chunk_size_targets = n_targets
                    n_chunks_targets = 1
                else:
                    n_chunks_targets = min(self.n_workers, n_targets)
                    chunk_size_targets = max(int(n_targets / self.n_workers), 1)
                    if self.chunk_size_states is None and n_chunks_states > 1:
                        while chunk_size_states * chunk_size_targets > n_targets:
                            n_chunks_states += 1
                            chunk_size_states = int(n_states / n_chunks_states)
            else:
                chunk_size_targets = min(n_targets, self.chunk_size_points)
                n_chunks_targets = max(int(n_targets / chunk_size_targets), 1)
            if int(n_targets / n_chunks_targets) > chunk_size_targets:
                n_chunks_targets += 1
                chunk_size_targets = int(n_targets / n_chunks_targets)
            chunk_sizes_targets = np.full(n_chunks_targets, chunk_size_targets)
            extra = n_targets - n_chunks_targets * chunk_size_targets
            if extra > 0:
                chunk_sizes_targets[-extra:] += 1

            s = np.sum(chunk_sizes_targets)
            assert s == n_targets, (
                f"Targets count mismatch: Expecting {n_targets}, chunks sum is {s}. Chunks: {[int(c) for c in chunk_sizes_targets]}"
            )

        chunk_sizes_states = np.full(n_chunks_states, chunk_size_states)
        extra = n_states - n_chunks_states * chunk_size_states
        if extra > 0:
            chunk_sizes_states[-extra:] += 1

        s = np.sum(chunk_sizes_states)
        assert s == n_states, (
            f"States count mismatch: Expecting {n_states}, chunks sum is {s}. Chunks: {[int(c) for c in chunk_sizes_states]}"
        )

        return chunk_sizes_states, chunk_sizes_targets

    def get_chunk_input_data(
        self,
        algo,
        model_data,
        farm_data,
        point_data,
        states_i0_i1,
        targets_i0_i1,
        out_vars,
        chunki_states,
        chunki_points,
        n_chunks_states,
        n_chunks_points,
    ):
        """
        Extracts the data for a single chunk calculation

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The algorithm object
        model_data: xarray.Dataset
            The initial model data
        farm_data: xarray.Dataset
            The initial farm data
        point_data: xarray.Dataset
            The initial point data
        states_i0_i1: tuple
            The (start, end) values of the states
        targets_i0_i1: tuple
            The (start, end) values of the targets
        out_vars: list of str
            Names of the output variables
        chunki_states: int
            The index of the states chunk
        chunki_points: int
            The index of the points chunk
        n_chunks_states: int
            The number of states chunks
        n_chunks_points: int
            The number of points chunks

        Returns
        -------
        data: tuple of foxes.core.Data
            The input data for the chunk calculation,
            either (mdata, fdata) or (mdata, fdata, tdata)

        """
        # prepare:
        i0_states, i1_states = states_i0_i1
        i0_targets, i1_targets = targets_i0_i1
        s_states = np.s_[i0_states:i1_states]
        s_targets = np.s_[i0_targets:i1_targets]

        # create mdata:
        mdata = MData.from_dataset(
            model_data,
            s_states=s_states,
            loop_dims=[FC.STATE],
            states_i0=i0_states,
            copy=True,
            chunki_states=chunki_states,
            chunki_points=chunki_points,
            n_chunks_states=n_chunks_states,
            n_chunks_points=n_chunks_points,
        )

        # create fdata:
        if farm_data is not None:
            fdata = FData.from_dataset(
                farm_data,
                mdata=mdata,
                s_states=s_states,
                callback=None,
                states_i0=i0_states,
                copy=True,
            )
        else:
            fdata = FData.from_mdata(
                mdata=mdata,
                states_i0=i0_states,
            )

        # create tdata:
        tdata = (
            TData.from_dataset(
                point_data,
                mdata=mdata,
                s_states=s_states,
                s_targets=s_targets,
                callback=None,
                states_i0=i0_states,
                copy=True,
            )
            if point_data is not None
            else None
        )

        return (mdata, fdata) if tdata is None else (mdata, fdata, tdata)

    def combine_results(
        self,
        algo,
        futures,
        model_data,
        out_vars,
        out_coords,
        n_chunks_states,
        n_chunks_targets,
        goal_data,
        iterative,
        write_nc=None,
        results=None,
    ):
        """
        Combines chunk results into final Dataset

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The algorithm object
        futures: dict
            The futures from the chunk calculations,
            key: (chunki_states, chunki_targets),
            value: future unpacking to (results, cstore)
        model_data: xarray.Dataset
            The initial model data
        out_vars: list of str
            Names of the output variables
        out_coords: list of str
            Names of the output coordinates
        n_chunks_states: int
            The number of states chunks
        n_chunks_targets: int
            The number of targets chunks
        goal_data: foxes.core.Data
            Either fdata or tdata
        iterative: bool
            Flag for use within the iterative algorithm
        write_nc: dict, optional
            Parameters for writing results to netCDF files, e.g.
            {'out_dir': 'results', 'base_name': 'calc_results',
            'ret_data': False, 'split': 1000}.

            The split parameter controls how the output is split:
            - 'chunks': one file per chunk (fastest method),
            - 'input': split according to sizes of multiple states input files,
            - int: split with this many states per file,
            - None: create a single output file.

            Use ret_data = False together with non-single file writing
            to avoid constructing the full Dataset in memory.
        results: dict, optional
            The evaluated futures

        Returns
        -------
        ds: xarray.Dataset, optional
            The final results dataset

        """
        assert (futures is not None and results is None) or (
            futures is None and results is not None
        ), f"{self.name}: Either futures or results must be provided, not both"

        self.print(f"{self.name}: Combining results", level=2)

        coords = {}
        vrb = max(self.verbosity - 1, 0)
        if FC.STATE in out_coords and FC.STATE in model_data.coords:
            coords[FC.STATE] = model_data[FC.STATE].to_numpy()

        # read parameters for file writing
        ret_data = True
        split_size = None
        write_on_fly = False
        write_from_ds = False
        keys = list(futures.keys()) if futures is not None else list(results.keys())
        if write_nc is not None and not (iterative and not algo.final_iteration):
            out_dir = get_output_path(write_nc.get("out_dir", "."))
            base_name = write_nc["base_name"]
            ret_data = write_nc.get("ret_data", False)
            split_mode = write_nc.get("split", None)
            out_dir.mkdir(parents=True, exist_ok=True)
            out_fpath = out_dir / (base_name + "_*.nc")
            if split_mode == "chunks":
                self.print(
                    f"{self.name}: Writing results to '{out_fpath}', using split = {split_mode}, ret_data = {ret_data}"
                )
            elif split_mode == "input":
                gen_size = algo.states.gen_states_split_size()
                split_size = next(gen_size)
            elif isinstance(split_mode, int):
                split_size = split_mode
            elif split_mode is None:
                split_size = None
            else:
                raise ValueError(
                    f"Invalid split mode '{split_mode}' in 'write_nc', expected 'chunks', 'input', int or None"
                )
            if split_size is None:
                out_fpath = out_dir / (base_name + ".nc")
            if split_mode != "chunks":
                write_on_fly = not ret_data and split_size is not None
                write_from_ds = not write_on_fly
                ret_data = write_nc.get("ret_data", write_from_ds)
                self.print(
                    f"{self.name}: Writing results to '{out_fpath}', using split = {split_mode}, on_fly = {write_on_fly}, ret_data = {ret_data}"
                )

        def _red_dims(data_vars):
            """Helper function for reducing unneccessary dimensions of data vars"""
            dvars = {}
            for v, (dims, d) in data_vars.items():
                if (
                    dims == (FC.STATE, FC.TURBINE)
                    and d.shape[1] == 1
                    and algo.n_turbines > 1
                ):
                    dvars[v] = ((FC.STATE,), d[:, 0])
                elif (
                    dims == (FC.STATE, FC.TARGET, FC.TPOINT)
                    and goal_data.sizes[FC.TARGET] > n_chunks_targets
                    and d.shape[1:] == (n_chunks_targets, 1)
                ):
                    dvars[v] = ((FC.STATE,), d[:, 0, 0])
                else:
                    dvars[v] = (dims, d)
            return dvars

        fcounter = 0

        def _write_parts_on_fly(data_vars, scount, wcount):
            """Helper function for writing chunked results to netCDF,
            without ever constructing the full Dataset in memory
            """
            nonlocal fcounter, split_size
            wfutures = []
            if split_size is not None:
                splits = min(split_size, algo.n_states - wcount)
                while algo.n_states - wcount > 0 and scount - wcount >= splits:
                    for v in data_vars.keys():
                        if len(data_vars[v][1]) > 1:
                            data_vars[v][1] = np.concatenate(data_vars[v][1], axis=0)
                        elif len(data_vars[v][1]) == 1:
                            data_vars[v][1] = data_vars[v][1][0]

                    dvars = {v: (d[0], d[1][:splits]) for v, d in data_vars.items()}
                    dvars = _red_dims(dvars)

                    crds = {v: d for v, d in coords.items()}
                    crds[FC.STATE] = coords[FC.STATE][wcount : wcount + splits]
                    if scount - wcount == splits:
                        for v in data_vars.keys():
                            data_vars[v][1] = []
                    else:
                        for v in data_vars.keys():
                            data_vars[v][1] = [data_vars[v][1][splits:]]

                    ds = Dataset(coords=crds, data_vars=dvars)
                    fpath = out_dir / f"{base_name}_{fcounter:04d}.nc"
                    args = (ds, fpath)
                    kwargs = dict(nc_engine=config.nc_engine, verbosity=vrb)
                    if len(futures) < self.n_workers:
                        future = self.submit(write_nc_file, *args, **kwargs)
                        wfutures.append(future)
                        del future
                    else:
                        write_nc_file(*args, **kwargs)
                    del ds, crds, dvars, args, kwargs

                    wcount += splits
                    fcounter += 1

                    if algo.n_states - wcount > 0:
                        if split_mode == "input":
                            try:
                                split_size = next(gen_size)
                            except StopIteration:
                                split_size = algo.n_states - wcount
                        splits = min(split_size, algo.n_states - wcount)

            return wcount, wfutures

        data_vars = {}

        def _get_res_vars(result, data_vars):
            """Helper function for extracting results variables"""
            if result is None:
                return []
            res_vars = list(result.keys())
            for v in out_vars:
                if v in res_vars:
                    data_vars[v] = [out_coords, []]
                else:
                    data_vars[v] = (goal_data[v].dims, goal_data[v].to_numpy())
            return res_vars

        if futures is not None:
            self.print(
                f"{self.name}: Computing {len(keys)} chunks using {self.n_workers} workers"
            )
            vlevel = 0
        else:
            self.print(f"{self.name}: Combining results from {len(keys)} chunks")
            vlevel = 1
        pbar = None
        if self.verbosity > vlevel and self.has_progress_bar:
            pbar = tqdm(total=len(keys))

        scount = 0
        wcount = 0
        wfutures = []
        res_vars = None
        pdone = -1
        if n_chunks_targets == 1:
            for key in keys:
                if futures is not None:
                    r, cstore = self.await_result(futures.pop(key))
                else:
                    r, cstore = results.pop(key)

                if res_vars is None:
                    res_vars = _get_res_vars(r, data_vars)
                if iterative:
                    for k, c in cstore.items():
                        if k in algo.chunk_store:
                            algo.chunk_store[k].update(c)
                        else:
                            algo.chunk_store[k] = c

                for v in res_vars:
                    if v in data_vars:
                        data_vars[v][1].append(r[v])

                if write_on_fly:
                    scount += r[res_vars[0]].shape[0]
                    wcount, ftrs = _write_parts_on_fly(data_vars, scount, wcount)
                    wfutures += ftrs
                    del ftrs
                del r, cstore

                if pbar is not None:
                    pbar.update()
                elif self.verbosity > vlevel and self.prints_progress:
                    pr = int(100 * (key[0] + 1) / len(keys))
                    if pr > pdone:
                        pdone = pr
                        print(
                            f"{self.name}: Completed {key[0] + 1} of {len(keys)} chunks, {pdone}%"
                        )
        else:
            pdone = -1
            counter = 0
            for chunki_states in range(n_chunks_states):
                tres = None
                for chunki_points in range(n_chunks_targets):
                    found = False
                    for key in keys:
                        if key == (chunki_states, chunki_points):
                            if futures is not None:
                                r, cstore = self.await_result(futures.pop(key))
                            else:
                                r, cstore = results.pop(key)

                            if res_vars is None:
                                res_vars = _get_res_vars(r, data_vars)
                            if tres is None:
                                tres = {v: [] for v in res_vars}

                            found = True
                            break
                    if not found:
                        raise KeyError(f"{self.name}: Missing future for key {key}")

                    for v in res_vars:
                        tres[v].append(r[v])
                    if iterative:
                        for k, c in cstore.items():
                            if k in algo.chunk_store:
                                algo.chunk_store[k].update(c)
                            else:
                                algo.chunk_store[k] = c
                    del r, cstore

                    counter += 1
                    if pbar is not None:
                        pbar.update()
                    elif (
                        self.verbosity > vlevel
                        and self.prints_progress
                        and len(keys) > 1
                    ):
                        pr = int(100 * counter / len(keys))
                        if pr > pdone:
                            pdone = pr
                            print(
                                f"{self.name}: Completed {counter} of {len(keys)} chunks, {pdone}%"
                            )

                found = False
                for v in res_vars:
                    if v in data_vars:
                        data_vars[v][1].append(np.concatenate(tres[v], axis=1))
                        if not found and write_on_fly:
                            scount += data_vars[v][1][-1].shape[0]
                        found = True
                del tres

                if write_on_fly:
                    wcount, ftrs = _write_parts_on_fly(data_vars, scount, wcount)
                    wfutures += ftrs
                    del ftrs
        for wf in wfutures:
            self.await_result(wf)
        del wfutures

        if pbar is not None:
            pbar.close()
        self.print(f"{self.name}: Completed all {len(keys)} chunks\n", level=vlevel + 1)

        # if not iterative or algo.final_iteration:
        #    algo.reset_chunk_store()

        if ret_data or write_from_ds:
            for v in res_vars:
                if v in data_vars:
                    if len(data_vars[v][1]) > 1:
                        data_vars[v][1] = np.concatenate(data_vars[v][1], axis=0)
                    elif len(data_vars[v][1]) == 1:
                        data_vars[v][1] = data_vars[v][1][0]

            dvars = _red_dims(data_vars)
            ds = Dataset(
                coords=coords,
                data_vars=dvars,
            )

            if write_from_ds:
                if split_size is None:
                    write_nc_file(
                        ds, out_fpath, nc_engine=config.nc_engine, verbosity=vrb
                    )
                else:
                    wcount = 0
                    fcounter = 0
                    wfutures = []
                    while wcount < algo.n_states:
                        splits = min(split_size, algo.n_states - wcount)
                        dssub = ds.isel({FC.STATE: slice(wcount, wcount + splits)})

                        fpath = out_dir / f"{base_name}_{fcounter:04d}.nc"
                        future = self.submit(
                            write_nc_file,
                            dssub,
                            fpath,
                            nc_engine=config.nc_engine,
                            verbosity=vrb,
                        )
                        wfutures.append(future)
                        del dssub, future

                        wcount += splits
                        fcounter += 1

                        if wcount < algo.n_states and split_mode == "input":
                            try:
                                split_size = next(gen_size)
                            except StopIteration:
                                split_size = algo.n_states - wcount

                    for wf in wfutures:
                        self.await_result(wf)
        if ret_data:
            return ds

    @abstractmethod
    def run_calculation(
        self,
        algo,
        model,
        model_data=None,
        farm_data=None,
        point_data=None,
    ):
        """
        Runs the model calculation

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The algorithm object
        model: foxes.core.DataCalcModel, optional
            The model that whose calculate function
            should be run
        model_data: xarray.Dataset
            The initial model data
        farm_data: xarray.Dataset, optional
            The initial farm data
        point_data: xarray.Dataset, optional
            The initial point data

        Returns
        -------
        results: xarray.Dataset
            The model results

        """
        n_states = model_data.sizes[FC.STATE]
        if point_data is None:
            self.print(
                f"{self.name}: Calculating {n_states} states for {algo.n_turbines} turbines"
            )
        else:
            self.print(
                f"{self.name}: Calculating data at {point_data.sizes[FC.TARGET]} points for {n_states} states"
            )
        if not self.initialized:
            raise ValueError(f"Engine '{self.name}' not initialized")
        if not model.initialized:
            raise ValueError(f"Model '{model.name}' not initialized")

    @classmethod
    def new(cls, engine_type, *args, **kwargs):
        """
        Run-time engine factory.

        Parameters
        ----------
        engine_type: str
            The selected derived class name
        args: tuple, optional
            Additional parameters for constructor
        kwargs: dict, optional
            Additional parameters for constructor

        """

        if engine_type is None:
            engine_type = "default"

        engine_type = dict(
            default="DefaultEngine",
            threads="ThreadsEngine",
            process="ProcessEngine",
            dask="DaskEngine",
            multiprocess="MultiprocessEngine",
            local_cluster="LocalClusterEngine",
            slurm_cluster="SlurmClusterEngine",
            mpi="MPIEngine",
            ray="RayEngine",
            numpy="NumpyEngine",
            single="SingleChunkEngine",
        ).get(engine_type, engine_type)

        return new_instance(cls, engine_type, *args, **kwargs)


def get_engine(error=True, default=True):
    """
    Gets the global calculation engine

    Parameters
    ----------
    error: bool
        Flag for raising ValueError if no
        engine is found
    default: bool or dict or Engine
        Set default engine if not set yet

    Returns
    -------
    engine: foxes.core.Engine
        The foxes calculation engine

    :group: core

    """
    engine = __global_engine_data__["engine"]
    if engine is None:
        if isinstance(default, dict):
            engine = Engine.new(**default)
            print(f"Selecting default engine '{engine}'")
            engine.initialize()
            return engine
        elif isinstance(default, Engine):
            print(f"Selecting default engine '{default}'")
            default.initialize()
            return default
        elif isinstance(default, bool) and default:
            engine = Engine.new(engine_type="DefaultEngine", verbosity=1)
            print(f"Selecting '{engine}'")
            engine.initialize()
            return engine
        elif error:
            raise ValueError("Engine not found.")
    return engine


def has_engine():
    """
    Flag that checks if engine has been set

    Returns
    -------
    flag: bool
        True if engine has been set

    :group: core

    """
    return __global_engine_data__["engine"] is not None


def reset_engine():
    """
    Resets the global calculation engine

    :group: core

    """
    engine = get_engine(error=False, default=False)
    if engine is not None:
        engine.finalize(type=None, value=None, traceback=None)



from abc import abstractmethod

from .model import Model


class DataCalcModel(Model):
    """
    Abstract base class for models
    that run calculation on xarray Dataset
    data.

    The calculations are run via xarray's
    `apply_ufunc` function, i.e., they run in
    parallel depending on the dask settings.

    For each individual data chunk the `calculate`
    function is called.

    :group: core

    """

    @abstractmethod
    def output_coords(self):
        """
        Gets the coordinates of all output arrays

        Returns
        -------
        dims: tuple of str
            The coordinates of all output arrays

        """
        pass

    @abstractmethod
    def calculate(self, algo, *data, **parameters):
        """
        The main model calculation.

        This function is executed on a single chunk of data,
        all computations should be based on numpy arrays.

        Parameters
        ----------
        algo: foxes.core.Algorithm
            The calculation algorithm
        data: tuple of foxes.core.Data, optional
            The input data, typically either (mdata, fdata) in
            the case of farm calculations, or (mdata, fdata, tdata)
            for point data calculations
        parameters: dict, optional
            The calculation parameters

        Returns
        -------
        results: dict
            The resulting data, keys: output variable str.
            Values: numpy.ndarray

        """
        pass



"""
Wake frame models.
"""

from .rotor_wd import RotorWD as RotorWD
from .streamlines import Streamlines2D as Streamlines2D
from .timelines import Timelines as Timelines
from .farm_order import FarmOrder as FarmOrder
from .seq_dynamic_wakes import SeqDynamicWakes as SeqDynamicWakes
from .dynamic_wakes import DynamicWakes as DynamicWakes



import time
import argparse
import matplotlib.pyplot as plt
import numpy as np

import foxes
import foxes.variables as FV


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-f",
        "--file_pattern",
        help="The search pattern for input *.nc files",
        default="wind_rotation.nc",
    )
    parser.add_argument(
        "-t",
        "--turbine_file",
        help="The P-ct-curve csv file (path or static)",
        default="NREL-5MW-D126-H90.csv",
    )
    parser.add_argument("-r", "--rotor", help="The rotor model", default="centre")
    parser.add_argument(
        "-p", "--pwakes", help="The partial wakes models", default="centre", nargs="+"
    )
    parser.add_argument(
        "-d", "--deflection", help="The wake deflection", default="no_deflection"
    )
    parser.add_argument(
        "-y",
        "--yawm",
        help="The uniform yaw misalignment value",
        type=float,
        default=None,
    )
    parser.add_argument(
        "-w",
        "--wakes",
        help="The wake models",
        default=["Jensen_linear_k007"],
        nargs="+",
    )
    parser.add_argument(
        "-wf", "--wake_frame", help="The wake frame choice", default="streamlines_100"
    )
    parser.add_argument(
        "-m", "--tmodels", help="The turbine models", default=[], nargs="+"
    )
    parser.add_argument(
        "-nt", "--n_turbines", help="The number of turbines", default=9, type=int
    )
    parser.add_argument(
        "-lm",
        "--load_mode",
        help="Dataset load mode",
        default="preload",
    )
    parser.add_argument("-e", "--engine", help="The engine", default="process")
    parser.add_argument(
        "-n", "--n_cpus", help="The number of cpus", default=None, type=int
    )
    parser.add_argument(
        "-c",
        "--chunksize_states",
        help="The chunk size for states",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-C",
        "--chunksize_points",
        help="The chunk size for points",
        default=None,
        type=int,
    )
    parser.add_argument(
        "-nf", "--nofig", help="Do not show figures", action="store_true"
    )
    args = parser.parse_args()

    mbook = foxes.models.ModelBook()
    ttype = foxes.models.turbine_types.PCtFile(args.turbine_file)
    mbook.turbine_types[ttype.name] = ttype

    # optionally set turbines in yaw:
    N = int(args.n_turbines**0.5)
    if args.yawm is None:
        ymodels = []
    else:
        yawm = np.zeros((1, N * N), dtype=np.float64)
        yawm[:, :N] = args.yawm
        mbook.turbine_models["set_yawm"] = foxes.models.turbine_models.SetFarmVars(
            pre_rotor=True
        )
        mbook.turbine_models["set_yawm"].add_var(FV.YAWM, yawm)
        ymodels = ["set_yawm"]

    states = foxes.input.states.FieldData(
        args.file_pattern,
        states_coord="state",
        x_coord="x",
        y_coord="y",
        h_coord="h",
        time_format=None,
        output_vars=[FV.WS, FV.WD, FV.TI, FV.RHO],
        var2ncvar={FV.WS: "ws", FV.WD: "wd"},
        fixed_vars={FV.RHO: 1.225, FV.TI: 0.1},
        load_mode=args.load_mode,
        interpn_pars=dict(bounds_error=False),
    )

    farm = foxes.WindFarm()
    foxes.input.farm_layout.add_grid(
        farm,
        xy_base=np.array([500.0, 500.0]),
        step_vectors=np.array([[500.0, 0], [0, 500.0]]),
        steps=(N, N),
        turbine_models=ymodels + args.tmodels + [ttype.name],
    )

    algo = foxes.algorithms.Downwind(
        farm,
        states,
        rotor_model=args.rotor,
        wake_models=args.wakes,
        wake_frame=args.wake_frame,
        wake_deflection=args.deflection,
        partial_wakes=args.pwakes,
        mbook=mbook,
    )

    with foxes.Engine.new(
        engine_type=args.engine,
        n_procs=args.n_cpus,
        chunk_size_states=args.chunksize_states,
        chunk_size_points=args.chunksize_points,
    ):
        time0 = time.time()
        farm_results = algo.calc_farm()
        time1 = time.time()

    print("\nCalc time =", time1 - time0, "\n")

    print(farm_results, "\n")

    fr = farm_results.to_dataframe()
    print(fr[[FV.X, FV.Y, FV.WD, FV.AMB_REWS, FV.REWS, FV.AMB_P, FV.P]])

    if not args.nofig:
        o = foxes.output.FlowPlots2D(algo, farm_results)
        for fig in o.gen_states_fig_xy(
            FV.WS,
            resolution=10,
            figsize=(8, 8),
            quiver_pars=dict(angles="xy", scale_units="xy", scale=0.07),
            quiver_n=15,
            xmin=0,
            xmax=2500,
            ymin=0,
            ymax=2500,
            rotor_color="red",
        ):
            plt.show()
            plt.close(fig)
